#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <bonobo.h>
#include <gnome.h>
#include <panel-applet.h>
#include <libgnomeui/gnome-help.h>
#include <libintl.h>
#include <syslog.h>
#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "common.h"

GThread *ui_thread;
GThread *tcp_thread;

#define GLADE_HOOKUP_OBJECT(component,widget,name) \
  g_object_set_data_full (G_OBJECT (component), name, \
    gtk_widget_ref (widget), (GDestroyNotify) gtk_widget_unref)

#define GLADE_HOOKUP_OBJECT_NO_REF(component,widget,name) \
  g_object_set_data (G_OBJECT (component), name, widget)

static void
g2ipmsg_applet_preferences_cb (BonoboUIComponent *uic,
                               GtkWidget        *applet,
                               const gchar       *cname)
{
  dbg_out("here\n");
  on_preferences1_activate(applet,NULL);
}
static void
g2ipmsg_applet_new_message_cb (BonoboUIComponent *uic,
                               GtkWidget        *applet,
                               const gchar       *cname)
{
  dbg_out("here\n");
  on_startBtn_clicked (applet,cname);
}
static void
g2ipmsg_applet_download_cb (BonoboUIComponent *uic,
                               GtkWidget        *applet,
                               const gchar       *cname)
{
  GtkWidget *window;

  dbg_out("here\n");

  window=create_downloadMonitor();
  g_assert(window);
  setup_download_window(window);
  gtk_widget_show(window);

}
static void
g2ipmsg_applet_remove_window_cb (BonoboUIComponent *uic,
                               GtkWidget        *applet,
                               const gchar       *cname)
{
  dbg_out("here\n");
  destroy_all_opened_windows();
}
static void
g2ipmsg_applet_foreground_window_cb (BonoboUIComponent *uic,
                               GtkWidget        *applet,
                               const gchar       *cname)
{
  dbg_out("here\n");
  present_all_displayed_windows();
}
static void
g2ipmsg_applet_about_cb (BonoboUIComponent *uic,
                               GtkWidget        *applet,
                               const gchar       *cname)
{
  dbg_out("here\n");
  ipmsg_show_about_dialog();
}
static gboolean
applet_event_button_press_event (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{

  dbg_out("buttonWin: button press: %d\n", event->button);
  switch (event->button) {
  case 1:
    on_startBtn_clicked (widget,user_data);
  default:	
    return FALSE;
    break;
  }
  return TRUE;
}

static const BonoboUIVerb g2ipmsg_applet_menu_verbs[] = {
        BONOBO_UI_UNSAFE_VERB ("G2ipmsgNewMessage",
                               g2ipmsg_applet_new_message_cb),
	BONOBO_UI_UNSAFE_VERB ("G2ipmsgDownLoadMonitor",
                               g2ipmsg_applet_download_cb),
	BONOBO_UI_UNSAFE_VERB ("G2ipmsgRemoveWindows",
                               g2ipmsg_applet_remove_window_cb),
	BONOBO_UI_UNSAFE_VERB ("G2ipmsgForegroundWindows",
			       g2ipmsg_applet_foreground_window_cb),
        BONOBO_UI_UNSAFE_VERB ("G2ipmsgPreferences",
                               g2ipmsg_applet_preferences_cb),
        BONOBO_UI_UNSAFE_VERB ("G2ipmsgAppletAbout",
                               g2ipmsg_applet_about_cb),
        BONOBO_UI_VERB_END
};

static GtkWidget*
applet_create_pixmap           (GtkWidget       *widget,
                                        const gchar     *pathname)
{
  GtkWidget *pixmap;

  if (!pathname || !pathname[0])
      return gtk_image_new ();

  dbg_out("pixmap file path:%s\n",pathname);

  pixmap = gtk_image_new_from_file (pathname);

  return pixmap;
}

static gboolean
g2ipmsg_applet_fill (PanelApplet *applet,
		   const gchar *iid,
		   gpointer     data)
{
	GdkPixbuf *icon_pixbuf;
	GtkWidget *startBtn;
	GtkWidget *image;
	GtkWidget *hbox;
	
        if (strcmp (iid, "OAFIID:G2ipmsgApplet") != 0)
		return FALSE;
	hbox=gtk_hbox_new(FALSE,1);
#if 1
	image = applet_create_pixmap (GTK_WIDGET(applet), GNOME_PIXMAPSDIR"/g2ipmsg/ipmsg.xpm");

	gtk_widget_show (image);
#endif
	gtk_box_pack_start(hbox,image,FALSE,TRUE,0);
	gtk_container_add (GTK_CONTAINER (applet), hbox);
	gtk_widget_show (hbox);
	gtk_widget_show_all (GTK_WIDGET (applet));

	gtk_widget_set_events (applet,GDK_BUTTON_PRESS_MASK);
	gtk_signal_connect (GTK_OBJECT (applet), "button_press_event",
		      GTK_SIGNAL_FUNC (applet_event_button_press_event),
		      NULL);


        panel_applet_setup_menu_from_file (PANEL_APPLET (applet),
                                           DATADIR,
                                           "g2ipmsg.xml",
                                           NULL,
                                           g2ipmsg_applet_menu_verbs,
                                           applet);

	/* Store pointers to all widgets, for use by lookup_widget(). */
	/* Note:They should have same name of initial window. */
	GLADE_HOOKUP_OBJECT_NO_REF (applet, applet, "GladeParentKey");
	GLADE_HOOKUP_OBJECT (applet, hbox, "startBtn");
	GLADE_HOOKUP_OBJECT (applet, image, "image3");



	hostinfo_init_hostinfo();
	init_ipmsg();
	tcp_thread=g_thread_create(ipmsg_tcp_server_thread,
			       PF_INET,
			       FALSE,
			       NULL);
	start_message_watcher(GTK_WIDGET(applet));
	dbg_out("applet init ok\n");

        return TRUE;
}


int
main (int argc, char *argv [])
{
  GnomeProgram *program;
  GOptionContext *context;
  int           retval;

  bindtextdomain (PACKAGE, GNOMELOCALEDIR);
  bind_textdomain_codeset (PACKAGE, "UTF-8");
  textdomain (PACKAGE);
  context = g_option_context_new ("");
  program = gnome_program_init ("GNOME2 IPMSG Applet", VERSION,
				LIBGNOMEUI_MODULE,
				argc, argv,	
				GNOME_PARAM_GOPTION_CONTEXT, context,
				GNOME_CLIENT_PARAM_SM_CONNECT, FALSE,
				GNOME_PROGRAM_STANDARD_PROPERTIES,
				GNOME_PARAM_APP_DATADIR, PACKAGE_DATA_DIR,NULL);				
  retval = panel_applet_factory_main ("OAFIID:G2ipmsgApplet_Factory", PANEL_TYPE_APPLET, g2ipmsg_applet_fill, NULL);
  g_object_unref (program);
  return retval;
}

