#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glib.h>
#include <gconf/gconf-client.h>
#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include "common.h"

#define GLADE_HOOKUP_OBJECT(component,widget,name) \
  g_object_set_data_full (G_OBJECT (component), name, \
    gtk_widget_ref (widget), (GDestroyNotify) gtk_widget_unref)

static GtkTargetEntry targetentries[] =
    {
      { "text/uri-list", 0, 0 },
    };

static void
setup_message_tree_view(GtkWidget *view) {
  GtkTreeViewColumn   *col;
  GtkCellRenderer     *renderer;
  
  /*
   * タイトル設定
   */
  /* --- Column #1 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("User"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 0);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #2 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("Group"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 1);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #3 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("Host"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 2);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #4 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("IP Address"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 3);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #5 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("LogOn"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 4);

  gtk_tree_view_column_set_resizable (col,TRUE);
}
static int
release_user_entry(GtkTreeView *view) {
  GtkTreePath *path;
  GtkTreeIter iter;
  GtkTreeModel *model;

  model = gtk_tree_view_get_model(view);
  
  if (!model)
    return -EINVAL;

  path=gtk_tree_path_new_first ();
  g_assert(path);

  if (gtk_tree_model_get_iter(model,&iter,path)) {
    while(gtk_list_store_remove(GTK_LIST_STORE(model), &iter))
      dbg_out("Remove OK\n");
  }
  gtk_tree_path_free (path);

  return 0;
}

static void
update_user_entry(GList *top,GtkWidget *view) {
  GtkListStore        *liststore;
  GtkTreeIter         toplevel;
  GtkLabel *usersNumlabel;
  userdb_t *current_user;
  int i;
  int users_count=0;
  gchar num_str[32];

  if ( (!top) || (!view) )
    return;

  if (!gtk_tree_view_get_column (GTK_TREE_VIEW(view),4))
    setup_message_tree_view(view);

  ipmsg_send_br_entry(udp_con,0);

  release_user_entry(GTK_TREE_VIEW(view));
  liststore = gtk_list_store_new(5,
                                 G_TYPE_STRING,
                                 G_TYPE_STRING,
                                 G_TYPE_STRING,
                                 G_TYPE_STRING,
                                 G_TYPE_STRING);
  for(i=0;(current_user=g_list_nth_data(top,i));++i) {
    _assert(current_user);
    
    dbg_out("Update\n");
    dbg_out("NickName: %s\n",current_user->nickname);
    dbg_out("Group: %s\n",current_user->group);
    dbg_out("User: %s\n",current_user->user);
    dbg_out("Host: %s\n",current_user->host);
    dbg_out("IP Addr: %s\n",current_user->ipaddr);
    /* Append a top level row and leave it empty */
    gtk_list_store_append(liststore, &toplevel);
    gtk_list_store_set(liststore, &toplevel,
		       0, current_user->nickname,
		       1,current_user->group,
		       2,current_user->host,
		       3,current_user->ipaddr,
		       4,current_user->user,
                     -1);
    ++users_count;
    fflush(stdout);
  }
  gtk_tree_view_set_model(GTK_TREE_VIEW(view), GTK_TREE_MODEL(liststore));

  g_object_unref(liststore); /* destroy model automatically with view */


  gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(view)),
                             GTK_SELECTION_MULTIPLE );
  /*
   * ユーザ数更新
   */
  usersNumlabel=(GtkLabel *)lookup_widget(GTK_WIDGET(view),"usersNumlabel");
  g_assert(usersNumlabel);
  snprintf(num_str,32,"%d",users_count);
  gtk_label_set_text(usersNumlabel,num_str);

  return;
}
static int
get_group_index(GtkComboBox *combobox,const gchar *groupname,int *index) {
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;
  int count=0;

  if ( (!combobox) || (!groupname) || (!index) )
    return -EINVAL;

  g_assert(combobox);
  model=gtk_combo_box_get_model(combobox);

  g_assert(model);
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while(valid) {
    gchar *str_data;

    gtk_tree_model_get (model, &iter, 
                          0, &str_data,
                          -1);
    dbg_out("group:%s\n",str_data);
    if (!strcmp(groupname,str_data)) {
      *index=count;
      dbg_out("find:%s(%d)\n",str_data,*index);
      g_free(str_data);
      return 0;
    }
    g_free(str_data);
    valid = gtk_tree_model_iter_next (model, &iter);
    ++count;
  }

  return -ESRCH;
}
static int
get_current_group_index(GtkComboBox *combobox) {
  int rc;
  int index=0;
  gchar *current_group;

  g_assert(combobox);

  current_group=(gchar *)hostinfo_refer_group_name();
  rc=get_group_index(combobox,current_group,&index);

  return index;
}
static int
updateConfigWindowGroups(GtkWidget *comboEntry) {
  GtkTreeModel *model;
  GtkTreePath *path;
  GtkTreeIter iter;
  gchar *active_text;
  int group_index=0;
  int rc;

  model=gtk_combo_box_get_model(GTK_COMBO_BOX(comboEntry));
  if (!model)
    return -EINVAL;

  path=gtk_tree_path_new_first ();
  g_assert(path);

  if (gtk_tree_model_get_iter(model,&iter,path)) {
    while(gtk_list_store_remove(GTK_LIST_STORE(model), &iter))
      dbg_out("Remove group\n");
  }
  gtk_tree_path_free (path);
  userdb_update_group_list(GTK_COMBO_BOX(comboEntry));

  active_text=gtk_combo_box_get_active_text(GTK_COMBO_BOX(comboEntry));
  if ( (active_text) &&   ((strlen(active_text))>0)) {
    rc=get_group_index(GTK_COMBO_BOX(comboEntry),active_text,&group_index);
    if (rc<0)
      gtk_combo_box_append_text (GTK_COMBO_BOX(comboEntry),active_text);
    else
      gtk_combo_box_set_active (GTK_COMBO_BOX(comboEntry),group_index);
  }

  if (active_text)
    g_free(active_text);

  return 0;
}
static void
sendmessage_get_attachment_part(long pktno,GtkWidget *editor,const char **ext_part){
  attach_file_block_t *afcb;
  GtkTreeIter  iter;
  GtkWidget    *view;
  GtkTreeModel *model;
  gboolean valid;
  gchar *string;
  int count;
  
  dbg_out("here\n");

  g_assert(ext_part);
  g_assert(*ext_part == NULL);

  if (create_attach_file_block(&afcb))
    return;

  view=lookup_widget(editor,"attachedFilesView");
  g_assert(view);

  model = gtk_tree_view_get_model(GTK_TREE_VIEW(view));
  g_assert(model);
  count=0;
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while(valid) {
      gchar *filepath;

      gtk_tree_model_get (model, &iter, 
                          0, &filepath,
                          -1);
      dbg_out("filepath:%s\n",filepath);
      add_attach_file(afcb,filepath);
      g_free(filepath);
      ++count;
      valid = gtk_tree_model_iter_next (model, &iter);
  }
  if ( (count > 0) && (!get_attach_file_extention(afcb,(const gchar **)&string)) ){
    add_upload_queue(pktno,afcb);
    dbg_out("Attach file string:%s\n",string);
    *ext_part=string;  
  }
}
static void
do_send(GtkTreeModel *model,
	       GtkTreePath *path,
	       GtkTreeIter *iter,
	       gpointer data){
  send_info_t *info;
  gchar *ipaddr;
  int pkt_no;
  int rc;
  GtkWidget *attachment_editor;
  gchar *ext_part=NULL;
  int lflags;

  info=(send_info_t *)data;
  gtk_tree_model_get (model, iter, 
		      3, &ipaddr,
		      -1);

  pkt_no=ipmsg_get_pkt_no();
  lflags=info->flags;
  dbg_out("Send to %s from ui\n",ipaddr);

  if ( lflags & IPMSG_FILEATTACHOPT){
    /*
     *アタッチメント
     */
    attachment_editor=info->attachment_editor;
    if (attachment_editor) {
      dbg_out("This message has attachment\n");
      ext_part=NULL;
      sendmessage_get_attachment_part(pkt_no,GTK_WIDGET(attachment_editor),(const char **)&ext_part);
      if (!ext_part) {
	lflags &= ~IPMSG_FILEATTACHOPT;
      }
    }
  }
  if ( lflags & IPMSG_FILEATTACHOPT){
    rc=ipmsg_send_send_msg(udp_con,ipaddr,lflags,pkt_no,info->msg,ext_part);
    if (!rc)
      ref_attach_file_block(pkt_no,ipaddr);
  } else {
    rc=ipmsg_send_send_msg(udp_con,ipaddr,lflags,pkt_no,info->msg,NULL);
  }
  if (rc)
      err_out("Can not send message to %s pktno=%d\n",ipaddr,pkt_no);
  if (ext_part)
    g_free(ext_part);
  g_free(ipaddr);
}
static int
has_dupulicated_string_in_cell(const char *string,GtkTreeView *treeview){
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;
  int rc;

  g_assert(treeview);
  if (!string)
    return -EINVAL;

  rc=-ESRCH;
  model = gtk_tree_view_get_model(treeview);
  for(valid = gtk_tree_model_get_iter_first (model, &iter);
      valid;
      valid = gtk_tree_model_iter_next (model, &iter)) {
	gchar *str_data;
	gtk_tree_model_get (model, &iter, 
                          0, &str_data,
                          -1);
	if (!strcmp(string,str_data)) {
	  dbg_out("Already exist:%s\n",str_data);
	  g_free(str_data);
	  rc=0;
	  break;
	}
	g_free(str_data);
  }
  return rc;
}
static void
onBroadcastSelectionChanged (GtkTreeSelection *sel, GtkListStore *liststore)
{
  GtkTreeIter  selected_row;
  GtkWidget    *view;
  GtkWidget    *remove_btn;
  
  view=GTK_WIDGET(gtk_tree_selection_get_tree_view(sel));
  g_assert(view);
  remove_btn=lookup_widget(GTK_WIDGET(view),"configRemoveBcastBtn");
  g_assert(remove_btn);
  g_assert(gtk_tree_selection_get_mode(sel) == GTK_SELECTION_SINGLE);

  if (gtk_tree_selection_get_selected(sel, NULL, &selected_row))
    {
      gtk_widget_set_sensitive(remove_btn, TRUE);
    }
  else
    {
      gtk_widget_set_sensitive(remove_btn, FALSE);
    }
}

void
on_app1_drag_data_received             (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


gboolean
on_app1_drag_drop                      (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        guint            time,
                                        gpointer         user_data)
{
  dbg_out("here\n");
  return FALSE;
}


gboolean
on_app1_destroy_event                  (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  dbg_out("here\n");
  return FALSE;
}


void
on_app1_destroy                        (GtkObject       *object,
                                        gpointer         user_data)
{
  GtkWidget *window;
  GtkWidget *attachment_editor;

  dbg_out("here\n");

  window=lookup_widget(GTK_WIDGET(object),"messageWindow");
  g_assert(window);

  attachment_editor = (GtkWidget*) g_object_get_data (G_OBJECT (object),
						      "attach_win");
  if (attachment_editor) 
    gtk_widget_destroy (attachment_editor);
  else
    dbg_out("No attachment editor\n");
}


void
on_new1_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


/* notice: This function may be called from applt, so we should not use
 * menuitem argument.
 */
void
on_preferences1_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *configWindow;
  GtkTreeView *broadcasts;
  GtkCellRenderer    *renderer;
  GtkTreeViewColumn  *col;
  GtkTreeSelection   *sel;
  GtkListStore       *liststore;

  dbg_out("here\n");
  configWindow=create_configWindow1 ();
  if (!configWindow) 
    return;

  broadcasts=GTK_TREE_VIEW(lookup_widget(configWindow,"treeview4"));
  g_assert(broadcasts);

  renderer = gtk_cell_renderer_text_new();
  if (!renderer)
    goto free_window;

  col = gtk_tree_view_column_new();
  if (!col)
    goto free_renderer;

  liststore = gtk_list_store_new(1, G_TYPE_STRING);
  if (!liststore)
    goto free_col;

  gtk_tree_view_set_model(GTK_TREE_VIEW(broadcasts), GTK_TREE_MODEL(liststore));
  g_object_unref(liststore);
  gtk_tree_view_column_pack_start(col, renderer, TRUE);
  gtk_tree_view_column_add_attribute(col, renderer, "text", 0);
  gtk_tree_view_column_set_title(col, _("Broad cast addresses"));
  gtk_tree_view_column_set_resizable (col,TRUE);

  gtk_tree_view_append_column(GTK_TREE_VIEW(broadcasts), col);

  sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(broadcasts));
  gtk_tree_selection_set_mode(sel, GTK_SELECTION_SINGLE);
  g_signal_connect(sel, "changed", 
		   G_CALLBACK(onBroadcastSelectionChanged),
		   liststore);
  

  gtk_widget_show (configWindow);

  return;

 free_window:
  gtk_widget_destroy(GTK_WIDGET(configWindow));
 free_col:
  gtk_object_destroy(GTK_OBJECT(col));
 free_renderer:
  gtk_object_destroy(GTK_OBJECT(renderer));
  return;
}


void
on_update1_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}

void
on_update2_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GList *users;
  GtkWidget *view;
  
  dbg_out("here\n");

  view=lookup_widget(GTK_WIDGET(menuitem),"messageUserTree");
  users=refer_user_list();
  update_user_entry(users,view);
}


void
on_about1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  ipmsg_show_about_dialog();
}


void
on_sendBtn_released                    (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_sendBtn_pressed                     (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}
void
on_sendBtn_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *top;
  GtkTextView *text_view;
  GtkTextIter start_iter,end_iter;
  GtkTextBuffer *buffer;
  GtkToggleButton *enclose_toggle;
  GtkTreeView *view;
  GtkTreeSelection *sel;
  GtkWidget *attachment_editor;
  gint text_line;
  gchar *string;
  gchar *to_string;
  int flags=0;
  send_info_t info;
  GList *path_list;
  GtkTreeModel *model;
  gint sender_count;

  dbg_out("here\n");

  top=lookup_widget(GTK_WIDGET(button),"messageWindow");
  text_view=GTK_TEXT_VIEW(lookup_widget(GTK_WIDGET(button),"textview1"));
  enclose_toggle=GTK_TOGGLE_BUTTON(lookup_widget(GTK_WIDGET(button),"encloseCheckBtn"));
  view=GTK_TREE_VIEW(lookup_widget(GTK_WIDGET(button),"messageUserTree"));
  attachment_editor = (GtkWidget*) g_object_get_data (G_OBJECT (top),
						      "attach_win");

  g_assert(view);
  g_assert(text_view);
  g_assert(enclose_toggle);
  g_assert(top);
    
  sel = gtk_tree_view_get_selection(view);

  sender_count=gtk_tree_selection_count_selected_rows(sel);
  if (!sender_count) {
    dbg_out("Not selected\n");
    return;
  }
  path_list=gtk_tree_selection_get_selected_rows(sel,&model);

  /* following codes assume gtk_text_view_get_buffer(text_view) does not
   * allocate new reference
   * as gtk manual says.
   */
  buffer=gtk_text_view_get_buffer(text_view);
  text_line=gtk_text_buffer_get_line_count(buffer);
  dbg_out("lines:%d\n",text_line);
  gtk_text_buffer_get_start_iter(buffer,&start_iter);
  gtk_text_buffer_get_end_iter(buffer,&end_iter);
  string=gtk_text_buffer_get_text(buffer,
				  &start_iter,
				  &end_iter,
				  FALSE);
  g_assert(string);
  dbg_out("string:%s\n",string);

  if (convert_string_ipmsg_proto(string,(const gchar **)&to_string)<0) {
      dbg_out("can not convert string :%s\n",string);
      goto free_out;
  }

  /*
   * 送信属性取得
   */
  flags=hostinfo_get_normal_send_flags();
  if (gtk_toggle_button_get_active(enclose_toggle))
    flags |= IPMSG_SECRETOPT;
  else
    flags &= ~IPMSG_SECRETOPT;

  info.flags=flags;
  info.msg=to_string;
  if (attachment_editor) {
    dbg_out("This message has attachment\n");
    info.attachment_editor=attachment_editor;
    info.flags |= IPMSG_FILEATTACHOPT;
  }

  gtk_tree_selection_selected_foreach(sel,do_send,&info);
  g_free(to_string);
  
 free_out:
  g_free(string);
  gtk_widget_destroy(top);

  return;
}  


void
on_encloseCheckBtn_toggled             (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configOpenCheckChkBtn_toggled       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configNonPopupCheckBtn_toggled      (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configNoSoundCheckBtn_toggled       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configEncloseEnableCheckBtn_toggled (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configCitationCheckBtn_toggled      (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configLoggingEnableCheckBtn_toggled (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configUserEnableCheckBtn_toggled    (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_logFileDialogBtn_pressed            (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *parent_window;
  GtkWidget *dialog;
  GtkWidget *file_entry;
  dbg_out("here\n");

  parent_window=lookup_widget(GTK_WIDGET(button),"configWindow1");

  dialog = gtk_file_chooser_dialog_new (_("LogFile"),
					GTK_WINDOW(parent_window),
					GTK_FILE_CHOOSER_ACTION_SAVE,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
					NULL);
  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
  if (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
  {
    char *filename;

    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    file_entry=lookup_widget(GTK_WIDGET(button),"entry1");
    gtk_entry_set_text(GTK_ENTRY(file_entry), filename); 
    g_free (filename);
  }
  gtk_widget_destroy (dialog);
}


void
on_logFileDialogBtn_released           (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_logFileDialogBtn_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_entry1_editing_done                 (GtkCellEditable *celleditable,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configAddBcastBtn_pressed           (GtkButton       *button,
                                        gpointer         user_data)
{
}


void
on_configAddBcastBtn_released          (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configAddBcastBtn_clicked           (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkEntry *entry;
  GtkTreeView *treeview;
  const gchar *txt;

  dbg_out("here\n");

  entry = GTK_ENTRY(lookup_widget(GTK_WIDGET(button),"entry3"));
  treeview = GTK_TREE_VIEW(lookup_widget(GTK_WIDGET(button),"treeview4"));

  txt = gtk_entry_get_text(GTK_ENTRY(entry));
  if (txt && *txt) {
      GtkTreeModel *model;
      GtkTreeIter   newrow;

      if (has_dupulicated_string_in_cell(txt,treeview)){
	model = gtk_tree_view_get_model(GTK_TREE_VIEW(treeview));
      
	gtk_list_store_append(GTK_LIST_STORE(model), &newrow);
      
	gtk_list_store_set(GTK_LIST_STORE(model), &newrow, 0, txt, -1);
      
	gtk_entry_set_text(GTK_ENTRY(entry), ""); /* clear entry */
      }
    }
}


void
on_configRemoveBcastBtn_pressed        (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configRemoveBcastBtn_released       (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configRemoveBcastBtn_clicked        (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkEntry *entry;
  GtkTreeView *treeview;
  GtkTreeSelection *sel;
  GtkTreeModel     *model;
  GtkTreeIter       selected_row;

  dbg_out("here\n");
  entry = GTK_ENTRY(lookup_widget(GTK_WIDGET(button),"entry3"));
  treeview = GTK_TREE_VIEW(lookup_widget(GTK_WIDGET(button),"treeview4"));

  sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));

  g_assert(gtk_tree_selection_get_mode(sel) == GTK_SELECTION_SINGLE);

  if (gtk_tree_selection_get_selected(sel, &model, &selected_row))
    {
      gchar *str_data;

      gtk_tree_model_get (model, &selected_row, 
                          0, &str_data,
                          -1);
      gtk_list_store_remove(GTK_LIST_STORE(model), &selected_row);
      dbg_out("Data:%s\n",str_data);
      gtk_entry_set_text(GTK_ENTRY(entry), str_data);
      g_free(str_data);
    }
  else
    {
      g_assert_not_reached();
    }

  
}


void
on_configApplySettingBtn_clicked       (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *top;
  GtkWidget *nickname_entry;
  GtkWidget *groupEntry;
  GtkWidget *logfile_entry;
  GtkWidget *configOpenCheckChkBtn;
  GtkWidget *configNonPopupCheckBtn;
  GtkWidget *configNoSoundCheckBtn;
  GtkWidget *configEncloseEnableCheckBtn;
  GtkWidget *configCitationCheckBtn;
  GtkWidget *enableLogToggle;
  GtkWidget *loginNameLoggingToggle;
  GtkWidget *logIPAddrToggle;
  GtkTreeView *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GSList *addr_top,*addr_list;
  gchar *string;
  gboolean valid;

  dbg_out("here\n");

  top=lookup_widget(GTK_WIDGET(button),"configWindow1");
  g_assert(top);
  groupEntry=lookup_widget(GTK_WIDGET(button),"comboboxentry1");
  g_assert(groupEntry);

  nickname_entry=lookup_widget(GTK_WIDGET(button),"entry2");
  g_assert(nickname_entry);

  logfile_entry=lookup_widget(GTK_WIDGET(button),"entry1");
  g_assert(logfile_entry);

  treeview = GTK_TREE_VIEW(lookup_widget(GTK_WIDGET(button),"treeview4"));
  g_assert(treeview);

  configOpenCheckChkBtn=lookup_widget(GTK_WIDGET(button),"configOpenCheckChkBtn");
  g_assert(configOpenCheckChkBtn);
  configNonPopupCheckBtn=lookup_widget(GTK_WIDGET(button),"configNonPopupCheckBtn");
  g_assert(configNonPopupCheckBtn);
  configNoSoundCheckBtn=lookup_widget(GTK_WIDGET(button),"configNoSoundCheckBtn");
  g_assert(configNoSoundCheckBtn);
  configEncloseEnableCheckBtn=lookup_widget(GTK_WIDGET(button),"configEncloseEnableCheckBtn");
  g_assert(configEncloseEnableCheckBtn);
  configCitationCheckBtn=lookup_widget(GTK_WIDGET(button),"configCitationCheckBtn");
  g_assert(configCitationCheckBtn);
  enableLogToggle=lookup_widget(GTK_WIDGET(button),"enableLogToggle");
  g_assert(enableLogToggle);
  loginNameLoggingToggle=lookup_widget(GTK_WIDGET(button),"loginNameLoggingToggle");
  g_assert(loginNameLoggingToggle);
  logIPAddrToggle=lookup_widget(GTK_WIDGET(button),"logIPAddrToggle");
  g_assert(logIPAddrToggle);

  hostinfo_set_nick_name(gtk_entry_get_text (GTK_ENTRY(nickname_entry)));

  string=gtk_combo_box_get_active_text(GTK_COMBO_BOX(groupEntry));
  hostinfo_set_group_name(string);
  g_free(string);

  hostinfo_set_ipmsg_logfile(gtk_entry_get_text(GTK_ENTRY(logfile_entry)));

  hostinfo_set_ipmsg_default_confirm(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(configOpenCheckChkBtn)));

  hostinfo_set_ipmsg_default_popup(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(configNonPopupCheckBtn)));

  hostinfo_set_ipmsg_default_sound(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(configNoSoundCheckBtn)));

  hostinfo_set_ipmsg_default_enclose(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(configEncloseEnableCheckBtn)));

  hostinfo_set_ipmsg_default_citation(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(configCitationCheckBtn)));

  hostinfo_set_ipmsg_enable_log(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(enableLogToggle)));

  hostinfo_set_ipmsg_logname_logging(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(loginNameLoggingToggle)));

  hostinfo_set_ipmsg_ipaddr_logging(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(logIPAddrToggle)));
 
  addr_top = NULL;
  model = gtk_tree_view_get_model(treeview);
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while(valid) {
      gchar *str_data;

      gtk_tree_model_get (model, &iter, 
                          0, &str_data,
                          -1);
      dbg_out("Addr:%s\n",str_data);
      addr_top=g_slist_append(addr_top,str_data);
      valid = gtk_tree_model_iter_next (model, &iter);
  }
  if (addr_top) {
    hostinfo_set_ipmsg_broadcast_list(addr_top);
    addr_list=addr_top;
    while(addr_list) {
      g_free(addr_list->data);
      addr_list=g_slist_next(addr_list);
    }
    g_slist_free(addr_top);
    g_assert(g_slist_length(addr_top)==1);
  }
  ipmsg_send_br_absence(udp_con,0);
  gtk_widget_destroy(top);
}


void
on_configApplySettingBtn_pressed       (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configApplySettingBtn_released      (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configCancelBtn_pressed             (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configCancelBtn_released            (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configCancelBtn_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *top;

  dbg_out("here\n");
  top=lookup_widget(GTK_WIDGET(button),"configWindow1");
  g_assert(top);
  gtk_widget_destroy(top);

}

GtkWidget *
internal_create_messageWindow(void){
  GtkWidget *messageWindow;
  GtkWidget *userview;
  GtkWidget *messageSeparator;
  GtkWidget *textview;
  gint width,height;

  messageWindow = create_messageWindow ();
  g_assert(messageWindow);

  userview = lookup_widget(messageWindow,"messageUserTree");
  g_assert(userview);

  messageSeparator = lookup_widget(messageWindow,"messageSeparator");
  g_assert(messageSeparator);

  textview = lookup_widget(messageWindow,"textview1");
  g_assert(textview);

  gtk_drag_dest_set(userview, GTK_DEST_DEFAULT_ALL, targetentries, 1,
		    GDK_ACTION_COPY|GDK_ACTION_MOVE|GDK_ACTION_LINK);


  gtk_drag_dest_set(messageSeparator, GTK_DEST_DEFAULT_ALL, targetentries, 1,
		    GDK_ACTION_COPY|GDK_ACTION_MOVE|GDK_ACTION_LINK);

  gtk_drag_dest_unset(textview);
  gtk_drag_dest_set(textview, GTK_DEST_DEFAULT_ALL, targetentries, 1,
		    GDK_ACTION_COPY|GDK_ACTION_MOVE|GDK_ACTION_LINK);

  if (!hostinfo_get_ipmsg_message_window_size(&width,&height)){
    dbg_out("Resize:(%d,%d)\n",width,height);
    gtk_window_resize (messageWindow,width,height);
  }
  return messageWindow;
}
void
on_startBtn_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  GList *users;
  GtkWidget *messageWindow;
  GtkWidget *treeview;
  /* We can not use button widget because this is called from applet */
  dbg_out("here\n");
  if (recv_windows_are_stored()) {
    show_stored_windows();
  } else {
    messageWindow = internal_create_messageWindow();
    g_assert(messageWindow);

    treeview=lookup_widget(messageWindow,"messageUserTree");
    g_assert(treeview);

    users=refer_user_list();
    update_user_entry(users,treeview);

    gtk_widget_show (messageWindow);
  }
}
static void
onAttachFileSelectionChanged (GtkTreeSelection *sel, GtkListStore *liststore)
{
	GtkTreeIter  selected_row;
	GtkWidget    *view;
	GtkWidget    *remove_btn;

	view=GTK_WIDGET(gtk_tree_selection_get_tree_view(sel));
	g_assert(view);
	remove_btn=lookup_widget(GTK_WIDGET(view),"AttachFileRemoveBtn");
	g_assert(remove_btn);
	g_assert(gtk_tree_selection_get_mode(sel) == GTK_SELECTION_SINGLE);

	if (gtk_tree_selection_get_selected(sel, NULL, &selected_row))
	{
		gtk_widget_set_sensitive(remove_btn, TRUE);
	}
	else
	{
		gtk_widget_set_sensitive(remove_btn, FALSE);
	}
}

static GtkWidget *
setup_attachment_editor(GtkWidget *mainwindow){
  GtkWidget           *attachment_editor;
  GtkWidget           *mainframe;
  GtkWidget           *entry;
  GtkTreeView         *view;
  GtkListStore        *liststore;
  GtkTreeViewColumn   *col;
  GtkCellRenderer     *renderer;
  GtkTreeSelection    *sel;

  dbg_out("here\n");
  attachment_editor = (GtkWidget*) g_object_get_data (G_OBJECT (mainwindow),
						      "attach_win");
  if (attachment_editor) {
    gtk_widget_grab_focus (attachment_editor);
    gtk_window_present (GTK_WINDOW(attachment_editor)); /* 前面移動  */
    return attachment_editor;
  }
  dbg_out("Create new editor\n");

  attachment_editor=create_attachFileEditor ();
  g_assert(attachment_editor);
  view=GTK_TREE_VIEW(lookup_widget(attachment_editor,"attachedFilesView"));  
  g_assert(view);

  liststore = gtk_list_store_new(3, 
				 G_TYPE_STRING,
				 G_TYPE_STRING,
				 G_TYPE_STRING);
  g_assert(liststore);

  gtk_tree_view_set_model(GTK_TREE_VIEW(view), GTK_TREE_MODEL(liststore));
  g_object_unref(liststore);

  /*
   * タイトル設定
   */
  /* --- Column #1 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("FilePath"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 0);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #2 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("FileSize"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 1);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #3 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("FileType"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 2);

  gtk_tree_view_column_set_resizable (col,TRUE);

  GLADE_HOOKUP_OBJECT (mainwindow, attachment_editor, "attach_win");
  GLADE_HOOKUP_OBJECT (attachment_editor,mainwindow,  "main_win");

  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
  gtk_tree_selection_set_mode(sel, GTK_SELECTION_SINGLE);
  g_signal_connect(sel, "changed", 
		   G_CALLBACK(onAttachFileSelectionChanged),
		   liststore);
  GTK_WIDGET_SET_FLAGS(attachment_editor,GTK_CAN_FOCUS);
  /*
   * Drop target setting
   */
  gtk_drag_dest_set(GTK_WIDGET(view), GTK_DEST_DEFAULT_ALL, targetentries, 1,
		    GDK_ACTION_COPY|GDK_ACTION_MOVE|GDK_ACTION_LINK);

  /*mainframe*/
  mainframe=lookup_widget(attachment_editor,"attachFileEditorMainFrame");
  g_assert(mainframe);
  gtk_drag_dest_set(mainframe, GTK_DEST_DEFAULT_ALL, targetentries, 1,
		    GDK_ACTION_COPY|GDK_ACTION_MOVE|GDK_ACTION_LINK);

  /*entry*/
  entry=lookup_widget(attachment_editor,"AttachFilePathEntry");
  g_assert(entry);
  gtk_drag_dest_set(entry, GTK_DEST_DEFAULT_ALL, targetentries, 1,
		    GDK_ACTION_COPY|GDK_ACTION_MOVE|GDK_ACTION_LINK);

  return attachment_editor;
}

static void
on_add_new_file(const gchar *filename,GtkWidget *parent_window,gboolean is_clear) {
  int rc;
  GtkWidget *file_entry;
  GtkTreeView *treeview;
  GtkTreeModel *model;
  GtkTreeIter   newrow;
  size_t num_size;
  time_t num_mtime;
  int    num_ftype;
  gchar str_size[64];
  gchar *str_ftype[]={
    _("UnKnown"),
    _("Regular"),
    _("Directory"),
    _("ReturnParent"),
    _("SymLink"),
    _("Character Device"),
    _("Block Device"),
    _("FIFO"),
    _("ResFork"),};

  dbg_out("here\n");

  file_entry=lookup_widget(parent_window,"AttachFilePathEntry");
  g_assert(file_entry);
  treeview = GTK_TREE_VIEW(lookup_widget(parent_window,"attachedFilesView"));
  g_assert(treeview);

  rc=has_dupulicated_string_in_cell(filename,treeview);
  if (!rc){
    dbg_out("Ignore duplicated entry %s\n",filename);
    return;
  }

  rc=get_file_info(filename,&num_size,&num_mtime,&num_ftype);
  if (rc<0) {
    dbg_out("Can not obtain file info %s %d\n",filename,num_ftype);
    return;
  }
  snprintf(str_size,62,"%d",num_size);
  str_size[63]=0;
  if (num_ftype==IPMSG_FILE_REGULAR) {/* 他の形式はサポート外  */
    model = gtk_tree_view_get_model(treeview);
    gtk_list_store_append(GTK_LIST_STORE(model), &newrow);
    gtk_list_store_set(GTK_LIST_STORE(model), &newrow, 
		       0, filename,
		       1, str_size,
		       2, str_ftype[num_ftype],
		       -1);
  }
  if (is_clear)
    gtk_entry_set_text(GTK_ENTRY(file_entry),"");  /* clear entry */
}

void
on_attachment1_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *window;
  GtkWidget *attachment_editor;

  window=lookup_widget(GTK_WIDGET(menuitem),"messageWindow");
  g_assert(window);
  attachment_editor=setup_attachment_editor(window);
  g_assert(attachment_editor);

  gtk_widget_show(attachment_editor);
}


void
on_quit1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  dbg_out("here\n");
  cleanup_message_watcher();
  gtk_main_quit();
}


void
on_entry2_activate                     (GtkEntry        *entry,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


gboolean
on_downloadWindow_destroy_event        (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  dbg_out("here\n");
  return FALSE;
}


void
on_clist4_click_column                 (GtkCList        *clist,
                                        gint             column,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_downloadSaveBtn_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_close_button1_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_configWindow_destroy                (GtkObject       *object,
                                        gpointer         user_data)
{

}


void
on_initialWindow_destroy               (GtkObject       *object,
                                        gpointer         user_data)
{

  dbg_out("here\n");
  on_quit1_activate((GtkMenuItem *)NULL,user_data);
}


void
on_downloadWindow_destroy              (GtkObject       *object,
                                        gpointer         user_data)
{
  GtkWidget *window;
  ipmsg_recvmsg_private_t *sender_info;
  ipmsg_private_data_t *priv;

  dbg_out("here\n");

  window=lookup_widget(GTK_WIDGET(object),"downloadWindow");
  g_assert(window);

  priv=(ipmsg_private_data_t *)lookup_widget(GTK_WIDGET(window),"senderInfo");
  g_assert(priv);
  sender_info=priv->data; 
  g_assert(sender_info->ipaddr);
  ipmsg_send_release_files(udp_con,sender_info->ipaddr,sender_info->pktno);
  remove_window_from_displaylist(window);
  /* privateは自動破棄される  */
}



void
on_messageWindow_drag_data_get         (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{

}


void
on_messageWindow_drag_end              (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gpointer         user_data)
{

}


void
on_downloadfilechooserdialog_destroy   (GtkObject       *object,
                                        gpointer         user_data)
{

}


GtkFileChooserConfirmation
on_downloadfilechooserdialog_confirm_overwrite
                                        (GtkFileChooser  *filechooser,
                                        gpointer         user_data)
{
  return TRUE;
}


void
on_downloadfilechooserdialog_current_folder_changed
                                        (GtkFileChooser  *filechooser,
                                        gpointer         user_data)
{

}


void
on_downloadfilechooserdialog_file_activated
                                        (GtkFileChooser  *filechooser,
                                        gpointer         user_data)
{

}


void
on_downloadfilechooserdialog_update_preview
                                        (GtkFileChooser  *filechooser,
                                        gpointer         user_data)
{

}


void
on_downloadfilechooserdialog_close     (GtkDialog       *dialog,
                                        gpointer         user_data)
{

}


void
on_downloadfilechooserdialog_response  (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{

}


void
on_logfileChooserdialog_destroy        (GtkObject       *object,
                                        gpointer         user_data)
{

}


void
on_logfileChooserdialog_close          (GtkDialog       *dialog,
                                        gpointer         user_data)
{

}


void
on_logfileChooserdialog_response       (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{

}


void
on_aboutdialog_destroy                 (GtkObject       *object,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_aboutdialog_close                   (GtkDialog       *dialog,
                                        gpointer         user_data)
{

}


void
on_aboutdialog_response                (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
  gtk_widget_destroy(GTK_WIDGET(dialog));
}


void
on_readNotifyDialog_close              (GtkDialog       *dialog,
                                        gpointer         user_data)
{

}


void
on_readNotifyDialog_response           (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{

}


void
on_readNotifyDialog_destroy            (GtkObject       *object,
                                        gpointer         user_data)
{

}


void
on_readNotifyDialogOKBtn_pressed       (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *top;

  top=lookup_widget(GTK_WIDGET(button),"readNotifyDialog");
  g_assert(top);
  gtk_widget_destroy(top);

}


gboolean
on_messageUserTree_select_cursor_row   (GtkTreeView     *treeview,
                                        gboolean         start_editing,
                                        gpointer         user_data)
{

  return FALSE;
}


gboolean
on_messageUserTree_drag_drop           (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        guint            time,
                                        gpointer         user_data)
{

  return FALSE;
}

void
on_messageWindow_show                  (GtkWidget       *widget,
                                        gpointer         user_data)
{
  GtkWidget *view;
  GtkToggleButton *enclose_toggle;
  GtkWidget *window;
  GtkWidget *textview;

  dbg_out("here\n");
  window=lookup_widget(widget,"messageWindow");
  enclose_toggle=GTK_TOGGLE_BUTTON(lookup_widget(widget,"encloseCheckBtn"));
  view=lookup_widget(widget,"messageUserTree");
  textview=lookup_widget(widget,"textview1");

  if (hostinfo_refer_ipmsg_default_enclose()) 
    gtk_toggle_button_set_active(enclose_toggle,TRUE);
  else
    gtk_toggle_button_set_active(enclose_toggle,FALSE);

}


void
on_viewWindow_destroy                  (GtkObject       *object,
                                        gpointer         user_data)
{
  GtkWidget *window;

  dbg_out("here\n");

  window=lookup_widget(GTK_WIDGET(object),"viewWindow");
  g_assert(window);
  remove_window_from_openlist(window);
  remove_window_from_displaylist(window);
}


void
on_viewWindow_show                     (GtkWidget       *widget,
                                        gpointer         user_data)
{

}


void
on_initialWindow_show                  (GtkWidget       *widget,
                                        gpointer         user_data)
{
  GtkWidget *startBtn;


  start_message_watcher(widget);

  startBtn=lookup_widget(widget,"startBtn");
  gtk_signal_connect (GTK_OBJECT (startBtn), "button_press_event",
		      GTK_SIGNAL_FUNC (on_init_win_event_button_press_event),
		      NULL);
}


void
on_viewwindowCloseBtn_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *top;

  top=lookup_widget(GTK_WIDGET(button),"viewWindow");
  g_assert(top);
  gtk_widget_destroy(top);
}

void
select_user_list_by_addr(GtkTreeView *treeview,const char *ipaddr) {
  GtkTreeSelection *sel;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;

  g_assert((treeview) && (ipaddr));

  sel=gtk_tree_view_get_selection(treeview);

  model = gtk_tree_view_get_model(treeview);
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while(valid) {
      gchar *tree_ipaddr;

      gtk_tree_model_get (model, &iter, 
                          3, &tree_ipaddr,
                          -1);
      dbg_out("Addr:%s\n",tree_ipaddr);
      if (!strcmp(tree_ipaddr,ipaddr)) {
	dbg_out("Select : %s\n",tree_ipaddr);
	gtk_tree_selection_select_iter(sel,&iter);
      }
      g_free(tree_ipaddr);
      valid = gtk_tree_model_iter_next (model, &iter);
  }
  
}
void
on_viewwindowReplyBtn_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *view;
  GtkWidget *messageWindow;
  GtkTextView *text_view;
  GtkTextIter start_iter,end_iter;
  GtkTextBuffer *buffer;
  GtkTextView *new_view;
  GtkTextBuffer *new_buffer;
  GtkWidget *cite_toggle;
  gint text_line;
  gchar *string;
  ipmsg_private_data_t *priv=NULL;
  ipmsg_recvmsg_private_t *sender_info;
  GtkWidget *messageUserTree;
  int i;

  dbg_out("here\n");

  messageWindow = create_messageWindow ();
  g_assert(messageWindow);
  messageUserTree=lookup_widget(messageWindow,"messageUserTree");
  g_assert(messageUserTree);

  view=lookup_widget(GTK_WIDGET(button),"viewWindow");
  text_view=GTK_TEXT_VIEW(lookup_widget(GTK_WIDGET(button),"viewwindowTextView"));
  cite_toggle=lookup_widget(GTK_WIDGET(button),"viewwindowCitationCheck");
  priv=(ipmsg_private_data_t *)lookup_widget(GTK_WIDGET(button),"senderInfo");

  g_assert(cite_toggle);
  g_assert(text_view);
  g_assert(view);
  g_assert(priv);

  IPMSG_ASSERT_PRIVATE(priv,IPMSG_PRIVATE_RECVMSG);
  sender_info=priv->data;
  dbg_out("sender_info:%p\n",sender_info);
  if (sender_info->ipaddr) {
    GList *users;

    users=refer_user_list();
    update_user_entry(users,messageUserTree);
    select_user_list_by_addr(GTK_TREE_VIEW(messageUserTree),sender_info->ipaddr);    
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cite_toggle))) {
    new_view=GTK_TEXT_VIEW(lookup_widget(messageWindow,"textview1"));
    new_buffer=gtk_text_view_get_buffer(new_view);
    
    g_assert(new_view);
    g_assert(new_buffer);

    /* following codes assume gtk_text_view_get_buffer(text_view) does not
     * allocate new reference
     * as gtk manual says.
     */
    buffer=gtk_text_view_get_buffer(text_view);
    text_line=gtk_text_buffer_get_line_count(buffer);
    dbg_out("lines:%d\n",text_line);
    for(i=0;i<text_line;++i) {
      
      gtk_text_buffer_get_iter_at_line(buffer,&start_iter,i);
      if (i<(text_line-1))
	gtk_text_buffer_get_iter_at_line(buffer,&end_iter,i+1);
      else
	gtk_text_buffer_get_end_iter(buffer,&end_iter);
      string=gtk_text_buffer_get_text(buffer,
				      &start_iter,
				      &end_iter,
				      FALSE);
      g_assert(string);
      gtk_text_buffer_insert_at_cursor(new_buffer,hostinfo_refer_ipmsg_cite_string(),-1);
      gtk_text_buffer_insert_at_cursor(new_buffer," ",-1);
      gtk_text_buffer_insert_at_cursor(new_buffer,string,-1);
      dbg_out("string:%s\n",string);
      g_free(string);
    }
  }

  gtk_widget_show (messageWindow);
  return;
}


void
on_viewwindowCitationCheck_toggled     (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


void
on_comboboxentry1_editing_done         (GtkCellEditable *celleditable,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_comboboxentry1_set_focus_child      (GtkContainer    *container,
                                        GtkWidget       *widget,
                                        gpointer         user_data)
{

}


void
on_entry1_activate                     (GtkEntry        *entry,
                                        gpointer         user_data)
{

}


void
on_entry3_activate                     (GtkEntry        *entry,
                                        gpointer         user_data)
{

}


void
on_entry3_editing_done                 (GtkCellEditable *celleditable,
                                        gpointer         user_data)
{

}

void
on_configWindow_show                   (GtkWidget       *widget,
                                        gpointer         user_data)
{
  GtkWidget *nickname_entry;
  GtkWidget *groupEntry;
  GtkWidget *logfile_entry;
  GtkWidget *configOpenCheckChkBtn;
  GtkWidget *configNonPopupCheckBtn;
  GtkWidget *configNoSoundCheckBtn;
  GtkWidget *configEncloseEnableCheckBtn;
  GtkWidget *configCitationCheckBtn;
  GtkWidget *enableLogToggle;
  GtkWidget *loginNameLoggingToggle;
  GtkWidget *logIPAddrToggle;
  GtkTreeView *treeview;
  GtkWidget *addr_entry;
  GtkWidget *configAddBcastBtn;
  GtkTreeIter  selected_row;
  GtkWidget *configRemoveBcastBtn;
  GtkTreeSelection *sel;
  gchar  *entry_string;
  GSList *addr_list,*addr_top;
  int group_index;
  int err_count;

  groupEntry=lookup_widget(widget,"comboboxentry1");
  g_assert(groupEntry);

  nickname_entry=lookup_widget(widget,"entry2");
  g_assert(nickname_entry);

  logfile_entry=lookup_widget(widget,"entry1");
  g_assert(logfile_entry);

  configOpenCheckChkBtn=lookup_widget(widget,"configOpenCheckChkBtn");
  g_assert(configOpenCheckChkBtn);
  configNonPopupCheckBtn=lookup_widget(widget,"configNonPopupCheckBtn");
  g_assert(configNonPopupCheckBtn);
  configNoSoundCheckBtn=lookup_widget(widget,"configNoSoundCheckBtn");
  g_assert(configNoSoundCheckBtn);
  configEncloseEnableCheckBtn=lookup_widget(widget,"configEncloseEnableCheckBtn");
  g_assert(configEncloseEnableCheckBtn);
  configCitationCheckBtn=lookup_widget(widget,"configCitationCheckBtn");
  g_assert(configCitationCheckBtn);
  enableLogToggle=lookup_widget(widget,"enableLogToggle");
  g_assert(enableLogToggle);
  loginNameLoggingToggle=lookup_widget(widget,"loginNameLoggingToggle");
  g_assert(loginNameLoggingToggle);
  logIPAddrToggle=lookup_widget(widget,"logIPAddrToggle");
  g_assert(logIPAddrToggle);

  treeview = GTK_TREE_VIEW(lookup_widget(widget,"treeview4"));
  g_assert(treeview);

  gtk_entry_set_text(GTK_ENTRY(nickname_entry), hostinfo_refer_nick_name());
  for(err_count=0,group_index=-1;
      ( (err_count<IPMSG_COMMON_MAX_RETRY) && (group_index<0) );
      ++err_count) {
    updateConfigWindowGroups(groupEntry);
    group_index=get_current_group_index(GTK_COMBO_BOX(groupEntry));
  }
  g_assert(group_index>=0);
  gtk_combo_box_set_active(GTK_COMBO_BOX(groupEntry),group_index);

  gtk_entry_set_text(GTK_ENTRY(logfile_entry), hostinfo_refer_ipmsg_logfile());

  if (hostinfo_refer_ipmsg_default_confirm())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configOpenCheckChkBtn),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configOpenCheckChkBtn),FALSE);

  if (hostinfo_refer_ipmsg_default_popup())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configNonPopupCheckBtn),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configNonPopupCheckBtn),FALSE);

  if (hostinfo_refer_ipmsg_default_sound())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configNoSoundCheckBtn),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configNoSoundCheckBtn),FALSE);

  if (hostinfo_refer_ipmsg_default_enclose()) 
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configEncloseEnableCheckBtn),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configEncloseEnableCheckBtn),FALSE);

  if (hostinfo_refer_ipmsg_default_citation())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configCitationCheckBtn),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(configCitationCheckBtn),FALSE);

  if (hostinfo_refer_ipmsg_enable_log())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enableLogToggle),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enableLogToggle),FALSE);

  if (hostinfo_refer_ipmsg_logname_logging())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(loginNameLoggingToggle),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(loginNameLoggingToggle),FALSE);

  if (hostinfo_refer_ipmsg_ipaddr_logging())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(logIPAddrToggle),TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(logIPAddrToggle),FALSE);


  addr_top=addr_list=hostinfo_get_ipmsg_broadcast_list();
  while(addr_list != NULL) {
      GtkTreeModel *model;
      GtkTreeIter   newrow;
      gchar *element = addr_list->data;

      model = gtk_tree_view_get_model(treeview);
      gtk_list_store_append(GTK_LIST_STORE(model), &newrow);
      gtk_list_store_set(GTK_LIST_STORE(model), &newrow, 0, element, -1);
      g_free(element);
      addr_list=g_slist_next(addr_list);
  }
  if (addr_top)
    g_slist_free(addr_top);

  /*
   * button state
   */
  configAddBcastBtn=lookup_widget(widget,"configAddBcastBtn");
  g_assert(configAddBcastBtn);
  addr_entry=lookup_widget(widget,"entry3");
  g_assert(addr_entry);

  entry_string=(char *)gtk_entry_get_text(GTK_ENTRY(addr_entry));
  if ( (entry_string) && (*entry_string) ) 
    gtk_widget_set_sensitive(configAddBcastBtn,TRUE);
  else
    gtk_widget_set_sensitive(configAddBcastBtn,FALSE);


  sel=gtk_tree_view_get_selection(treeview);
  configRemoveBcastBtn=lookup_widget(widget,"configRemoveBcastBtn");
  g_assert(configRemoveBcastBtn);
  g_assert(gtk_tree_selection_get_mode(sel) == GTK_SELECTION_SINGLE);

  if (gtk_tree_selection_get_selected(sel, NULL, &selected_row))
    {
      gtk_widget_set_sensitive(configRemoveBcastBtn, TRUE);
    }
  else
    {
      gtk_widget_set_sensitive(configRemoveBcastBtn, FALSE);
    }

}

void
on_configWindow_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{

}

void
update_one_group_entry(gpointer data,GtkComboBox *user_data) {
  userdb_t *current_user;
  GtkWidget *comboEntry=GTK_WIDGET(user_data);
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;

  if ( (!data) || (!user_data) )
    return;

  current_user=(userdb_t *)data;

  dbg_out("NickName: %s\n",current_user->nickname);
  dbg_out("Group: %s\n",current_user->group);
  dbg_out("User: %s\n",current_user->user);
  dbg_out("Host: %s\n",current_user->host);
  dbg_out("IP Addr: %s\n",current_user->ipaddr);

  model=gtk_combo_box_get_model(GTK_COMBO_BOX(comboEntry));
  if (!model)
    return;

  valid = gtk_tree_model_get_iter_first (model, &iter);

  while (valid) {
      gchar *str_data;
      /* Make sure you terminate calls to gtk_tree_model_get()
       * with a '-1' value
       */
      gtk_tree_model_get (model, &iter, 
                          0, &str_data,
                          -1);
	dbg_out("check: %s\n",str_data);
      if (!strcmp(current_user->group,str_data)) {
	dbg_out("found: %s\n",current_user->group);
	g_free(str_data);
	return; /* Already appended */
      }
      g_free(str_data);
      valid = gtk_tree_model_iter_next (model, &iter);
    }

  gtk_combo_box_append_text(GTK_COMBO_BOX(comboEntry),current_user->group);

  return;
}
void
on_configWindowAddGroupBtn_clicked     (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *comboEntry;

  ipmsg_send_br_absence(udp_con,0);
  comboEntry=lookup_widget(GTK_WIDGET(button),"comboboxentry1");
  g_assert(comboEntry);
  updateConfigWindowGroups(comboEntry);
}


void
on_loginNameLoggingToggle_toggled      (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


void
on_logIPAddrToggle_toggled             (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


void
on_enableLogToggle_toggled             (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


void
on_sendFailDialog_show                 (GtkWidget       *widget,
                                        gpointer         user_data)
{

}


void
on_sendFailDialog_destroy              (GtkObject       *object,
                                        gpointer         user_data)
{

}


void
on_sendFailDialog_response             (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
  dbg_out("Response:%d\n",response_id);
}


void
on_sendFailDialog_close                (GtkDialog       *dialog,
                                        gpointer         user_data)
{

}


void
on_cancelbutton1_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_okbutton1_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *dialog;
  dialog=lookup_widget(GTK_WIDGET(button),"sendFailDialog");
  gtk_dialog_response(GTK_DIALOG(dialog),GTK_RESPONSE_OK);
}

static int
handle_attachment_drag_data(GtkSelectionData *data,GtkWidget *attachment_editor){
  GList *files,*list;

  if ( (!data) || (!(data->data)) || (!attachment_editor) )
    return -EINVAL;

  files= gnome_vfs_uri_list_parse((gchar *)data->data);
  for(list = files; list != NULL; list = g_list_next(list) ) {
    dbg_out("draged file: %s\n",gnome_vfs_uri_get_path(list->data));
    on_add_new_file(gnome_vfs_uri_get_path(list->data),attachment_editor,FALSE);
  }
  gnome_vfs_uri_list_free(files);

  return 0;
}
static int 
on_message_window_drag_data_received(GtkWidget *widget,GtkSelectionData *data){
  GtkWidget *window;
  GtkWidget *attachment_editor;
  int rc;

  if (!data)
    return -EINVAL;

  dbg_out("here:data %s\n",(char *)data->data);  

  window=lookup_widget(widget,"messageWindow");
  g_assert(window);
  attachment_editor=setup_attachment_editor(window);
  g_assert(attachment_editor);

  gtk_widget_show(attachment_editor);

  rc=handle_attachment_drag_data(data,attachment_editor);

  if (rc<0)
    err_out("Can not handle drag data %s (%d)\n",strerror(-rc),-rc);

  return rc;
}

void
on_messageUserTree_drag_data_received  (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{
  g_assert(widget);
  g_assert(data);

  dbg_out("here:data %s\n",(char *)data->data);  
  on_message_window_drag_data_received(widget,data);
}



void
on_close1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *top;

  dbg_out("here\n");
  top=lookup_widget(GTK_WIDGET(menuitem),"messageWindow");
  g_assert(top);
  gtk_widget_destroy(top);
}


void
on_attachFileEditor_destroy            (GtkObject       *object,
                                        gpointer         user_data)
{
  GtkWidget *mainwindow;
  GtkWidget *attachment_editor;

  dbg_out("here\n");
  mainwindow = (GtkWidget*) g_object_get_data (G_OBJECT (object),
						      "main_win");
  g_assert(mainwindow);
  dbg_out("Remove Assosiation\n");
  attachment_editor =  (GtkWidget*) g_object_get_data (G_OBJECT (mainwindow),
						      "attach_win");
  g_assert(attachment_editor);
  g_object_set_data (G_OBJECT (attachment_editor),
		     "main_win",
		     NULL);
  g_object_set_data (G_OBJECT (mainwindow),
		     "attach_win",
		     NULL);
  attachment_editor =  (GtkWidget*) g_object_get_data (G_OBJECT (mainwindow),
						      "attach_win");
  g_assert(!attachment_editor);
}


void
on_attachFileEditor_show               (GtkWidget       *widget,
                                        gpointer         user_data)
{
  GtkWidget *attachment_editor;
  GtkWidget *add_btn;
  GtkWidget *remove_btn;
  GtkWidget *view;
  GtkWidget *entry;
  GtkTreeSelection *sel;
  GtkTreeModel     *model;
  GtkTreeIter  selected_row;
  size_t size;
  time_t mtime;
  int type;
  int rc;
  gchar *filepath;

  dbg_out("here\n");

  /*
   * setup button/entry
   */
  attachment_editor=lookup_widget(widget,"attachFileEditor");
  g_assert(attachment_editor);

  add_btn=lookup_widget(GTK_WIDGET(attachment_editor),"AttachFIleAddBtn");
  g_assert(add_btn);
  remove_btn=lookup_widget(GTK_WIDGET(attachment_editor),"AttachFileRemoveBtn");
  g_assert(remove_btn);
  entry=lookup_widget(GTK_WIDGET(attachment_editor),"AttachFilePathEntry");
  g_assert(entry);

  filepath=(char *)gtk_entry_get_text(GTK_ENTRY(entry));
  rc=get_file_info(filepath,&size,&mtime,&type);
  if ( (rc) || (type != IPMSG_FILE_REGULAR) )
    gtk_widget_set_sensitive(add_btn,FALSE);
  else
    gtk_widget_set_sensitive(add_btn,TRUE);

  view=lookup_widget(GTK_WIDGET(attachment_editor),"attachedFilesView");
  g_assert(GTK_TREE_VIEW(view));

  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
  g_assert(gtk_tree_selection_get_mode(sel) == GTK_SELECTION_SINGLE);

  if (gtk_tree_selection_get_selected(sel, &model, &selected_row))
    gtk_widget_set_sensitive(remove_btn,TRUE);
  else
    gtk_widget_set_sensitive(remove_btn,FALSE);
}


void
on_AttachFilePathEntry_activate        (GtkEntry        *entry,
                                        gpointer         user_data)
{
  GtkWidget *parent_window;
  char *filename;

  dbg_out("here\n");

  parent_window=lookup_widget(GTK_WIDGET(entry),"attachFileEditor");
  filename=(char *)gtk_entry_get_text(entry); 
}

void
on_AttachFileFileChooseBtn_clicked     (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *parent_window;
  GtkWidget *dialog;
  GtkWidget *file_entry;
  dbg_out("here\n");

  parent_window=lookup_widget(GTK_WIDGET(button),"attachFileEditor");

  dialog = gtk_file_chooser_dialog_new (_("AttachmentFile"),
					GTK_WINDOW(parent_window),
					GTK_FILE_CHOOSER_ACTION_OPEN,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					NULL);
  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
  if (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
  {
    char *filename;

    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    file_entry=lookup_widget(GTK_WIDGET(button),"AttachFilePathEntry");
    gtk_entry_set_text(GTK_ENTRY(file_entry), filename); 
    g_free (filename);
  }
  gtk_widget_destroy (dialog);
}
void
on_AttachFIleAddBtn_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *parent_window;
  GtkWidget *file_entry;
  char *filename;

  dbg_out("here\n");

  parent_window=lookup_widget(GTK_WIDGET(button),"attachFileEditor");
  file_entry=lookup_widget(GTK_WIDGET(button),"AttachFilePathEntry");
  filename=(char *)gtk_entry_get_text(GTK_ENTRY(file_entry)); 

  on_add_new_file(filename,parent_window,TRUE);
}


void
on_AttachFileRemoveBtn_clicked         (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkTreeIter  selected_row;
  GtkWidget    *view;
  GtkTreeSelection *sel;
  GtkTreeModel *model;
  GtkWidget    *remove_btn;
  GtkWidget    *file_entry;

  dbg_out("here\n");

  view=lookup_widget(GTK_WIDGET(button),"attachedFilesView");
  g_assert(view);
  remove_btn=lookup_widget(GTK_WIDGET(button),"AttachFileRemoveBtn");
  g_assert(remove_btn);  
  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
  g_assert(sel);  
  file_entry=lookup_widget(GTK_WIDGET(button),"AttachFilePathEntry");
  g_assert(file_entry);
  g_assert(gtk_tree_selection_get_mode(sel) == GTK_SELECTION_SINGLE);

  if (gtk_tree_selection_get_selected(sel, &model, &selected_row))
    {
      gchar *tree_path;

      gtk_tree_model_get (model, &selected_row, 
                          0, &tree_path,
                          -1);
      dbg_out("Selected: %s\n",tree_path);
      gtk_entry_set_text(GTK_ENTRY(file_entry), tree_path);       
      g_free(tree_path);
      gtk_list_store_remove(GTK_LIST_STORE(model), &selected_row);
      gtk_widget_set_sensitive(remove_btn, FALSE);
    }
}

static int 
on_attach_window_drag_data_received(GtkWidget *widget,GtkSelectionData *data){
  GtkWidget *attachment_editor;
  int rc;

  if (!data)
    return -EINVAL;

  dbg_out("here:data %s\n",(char *)data->data);  


  attachment_editor=lookup_widget(widget,"attachFileEditor");
  g_assert(attachment_editor);

  gtk_widget_show(attachment_editor);

  rc=handle_attachment_drag_data(data,attachment_editor);
  if (rc<0)
    err_out("Can not handle drag data %s (%d)\n",strerror(-rc),-rc);

  return rc;
}

void
on_attachedFilesView_drag_data_received
                                        (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{
  int rc;

  dbg_out("here\n");

  if (!data)
    return;

  rc=on_attach_window_drag_data_received(widget,data);

  if (rc<0)
    err_out("Can not handle drag data %s (%d)\n",strerror(-rc),-rc);

  return;
}


void
on_AttachFilePathEntry_changed         (GtkEditable     *editable,
                                        gpointer         user_data)
{
  GtkWidget *file_entry;
  GtkWidget *btn;
  gchar *filepath;
  size_t size;
  time_t mtime;
  int type;
  int rc;

  btn=lookup_widget(GTK_WIDGET(editable),"AttachFIleAddBtn");
  file_entry=lookup_widget(GTK_WIDGET(editable),"AttachFilePathEntry");
  filepath=(char *)gtk_entry_get_text(GTK_ENTRY(file_entry));
  rc=get_file_info(filepath,&size,&mtime,&type);
  if ( (rc) || (type != IPMSG_FILE_REGULAR) )
    gtk_widget_set_sensitive(btn,FALSE);
  else
    gtk_widget_set_sensitive(btn,TRUE);
}

static int
do_download(GtkTreeModel *model,
	       GtkTreePath *path,
	       GtkTreeIter *iter,
	       gpointer data){
  int rc;
  int sock;
  int fd;
  GtkWidget *window;
  GtkWidget *entry;
  ipmsg_recvmsg_private_t *sender_info;
  ipmsg_private_data_t *priv;
  long pkt_no;
  int fileid;
  char buff[_MSG_BUF_SIZE];
  ssize_t recv_len;
  ssize_t len;
  ssize_t soc_remains;
  ssize_t write_len;
  size_t total_write;
  size_t file_remains;
  char filepath[PATH_MAX];
  char *filename;
  char *dirname;
  char *req_message=NULL;

  window=GTK_WIDGET(data);
  g_assert(window);
  entry=lookup_widget(GTK_WIDGET(window),"DownLoadDirectoryEntry");
  g_assert(entry);

  priv=(ipmsg_private_data_t *)lookup_widget(GTK_WIDGET(window),"senderInfo");
  g_assert(priv);
  sender_info=priv->data; 
  g_assert(sender_info->ipaddr);

  pkt_no=sender_info->pktno;

  gtk_tree_model_get (model, iter, 
		      0, &fileid,
		      1, &filename,
		      3, &file_remains,
		      -1);
  dirname=(char *)gtk_entry_get_text(GTK_ENTRY(entry));
  snprintf(filepath,PATH_MAX-1,"%s/%s",dirname,filename);
  filepath[PATH_MAX-1]='\0';
  g_free(filename);

  dbg_out("download %d th element of %ld from %s into %s\n",fileid,pkt_no,sender_info->ipaddr,filepath);
  fd=open(filepath,O_RDWR|O_EXCL|O_CREAT,S_IRWXU|S_IRWXG|S_IRWXO);
  if (fd<0) {
    if (errno == EEXIST) {
      GtkWidget *dialog=create_ipmsgDownloadOverWrite ();
      GtkWidget *label;
      gint result;

      g_assert(dialog);
      label=GTK_WIDGET(lookup_widget(dialog,"overwriteFileNameLabel"));
      gtk_label_set_text(GTK_DIALOG(label),filepath);
      result=gtk_dialog_run (GTK_DIALOG (dialog));
      switch (result) {
      case GTK_RESPONSE_OK:
	dbg_out("Accept for overwrite\n");
	fd=open(filepath,O_RDWR|O_CREAT,S_IRWXU|S_IRWXG|S_IRWXO);
	if (fd<0) {
	  err_out("Can not open file:%s (%d)\n",strerror(errno),errno);
	  gtk_widget_destroy (dialog);
	  rc=-errno;
	  goto no_close_out;
	}
	break;
      default:
	err_out("response:%d Can not open file:%s (%d)\n",result,strerror(errno),errno);
	gtk_widget_destroy (dialog);
	rc=-ENOENT;
	goto no_close_out;
	break;
      }
      gtk_widget_destroy (dialog);
    } else{
      err_out("Can not open file:%s (%d)\n",strerror(errno),errno);
      rc=-errno;
      goto no_close_out;
    }
  }

  rc=tcp_setup_client(PF_INET,sender_info->ipaddr,hostinfo_refer_ipmsg_port(),&sock);
  if (rc<0) {
    rc=-errno;
    goto file_close_out;
  }
  memset(buff,0,TCP_FILE_BUFSIZ);
  snprintf(buff,TCP_FILE_BUFSIZ-1,"%x:%x:0",(unsigned int)pkt_no,(unsigned int)fileid);
  buff[TCP_FILE_BUFSIZ-1]='\0';

  dbg_out("request:%s\n",buff);

  rc=ipmsg_construct_getfile_message(buff,0,(size_t *)&len,&req_message,ipmsg_get_pkt_no());
  if (rc<0) {
    rc=-errno;
    goto file_close_out;
  }
  dbg_out("request message:%s\n",req_message);

  rc=wait_socket(sock,WAIT_FOR_WRITE,TCP_SELECT_SEC);
  if (rc<0) {
    err_out("Can not write for socket %s (%d)\n",strerror(errno),errno);
    free(req_message);
    goto err_io;
  }  

  soc_remains=len;
  while(soc_remains > 0) {
    write_len=send(sock, req_message,len, MSG_NOSIGNAL);
    if (write_len < 0) {
      err_out("Can not write socket :%s (%d)\n",strerror(errno),errno);
      free(req_message);
      rc=-errno;
      goto err_io;
    }
    soc_remains -= write_len;
    dbg_out("Write remains:%d\n",soc_remains);
  }
  free(req_message);

  total_write=0;
  dbg_out("Try to read %d byte total\n", file_remains);
  /*
   *読み取り
   */
  while(file_remains>0) {
    rc=wait_socket(sock,WAIT_FOR_READ,TCP_SELECT_SEC);
    if (rc<0) {
      err_out("Can not read socket :%s (%d)\n",strerror(errno),errno);
      goto err_io;
    }
    memset(buff, 0, sizeof(buff));
    errno=0;
    recv_len=recv(sock,buff,_MSG_BUF_SIZE,(MSG_PEEK));
    if (!recv_len) {
      rc=-errno;
      goto err_io; /* err or end */
    }
    if (recv_len<=0) {
      if (errno) 
	err_out("Can not peek message %s(errno:%d)\n",strerror(errno),errno);
      rc=-errno;
      goto err_io; /* err or end */
    }


    dbg_out("tcp peek read:%d\n",recv_len);
    if (file_remains<recv_len) {
      recv_len = file_remains;
      dbg_out("fixup size:%d\n",recv_len);
    }
    errno=0;
    recv_len=recv(sock,buff,recv_len,MSG_DONTWAIT);
    if (recv_len<=0) {
      if (errno) 
	err_out("Can not recv message %s(errno:%d)\n",strerror(errno),errno);
      rc=-errno;
      goto err_io; /* err or end */
    }
    file_remains -= recv_len;
    dbg_out("tcp read:%d(%d remains)\n",recv_len,file_remains);
    write_len=write(fd,buff,recv_len);
    if (write_len<0) {
      err_out("Can not write file %s(errno:%d)\n",strerror(errno),errno);
      rc=-errno;
      goto err_io; /* err or end */
    }
    total_write += write_len;
    if (!GTK_IS_LIST_STORE (model)) {
      war_out("Download window is destroyed.\n");
      rc=-EFAULT; /*  仮にEFAULTを返す  */
      goto err_io; /* window destory */
    }
    gtk_list_store_set(GTK_LIST_STORE(model), iter, 
		       2, total_write,
		       -1);
    dbg_out("tcp write file :%d\n",total_write);
    while (gtk_events_pending ())
      gtk_main_iteration ();
  }
  /* 0バイトファイル対策のためもう一度書き込む  */
  gtk_list_store_set(GTK_LIST_STORE(model), iter, 
		     2, total_write,
		     -1);
  rc=0;
 err_io:
  close(sock); /* 端点のクローズは, 受信側で行うのがipmsgの仕様  */
 file_close_out:
  close(fd);
 no_close_out:
  return rc;
}

void
on_DownLoadOKBtn_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  int rc;
  GtkWidget *window;
  ipmsg_recvmsg_private_t *sender_info;
  ipmsg_private_data_t *priv;
  GtkWidget *dirbtn;
  GtkWidget    *view;
  GtkTreeSelection *sel;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GList *pathes,*node;
  gboolean noneed_to_destory;


  dbg_out("here:\n");
  window=lookup_widget(GTK_WIDGET(button),"downloadWindow");
  g_assert(window);
  dirbtn=lookup_widget(GTK_WIDGET(window),"DownLoadOpenBtn");
  g_assert(dirbtn);

  priv=(ipmsg_private_data_t *)lookup_widget(GTK_WIDGET(window),"senderInfo");
  g_assert(priv);
  sender_info=priv->data; 
  g_assert(sender_info->ipaddr);

  gtk_widget_set_sensitive(dirbtn,FALSE);
  view=lookup_widget(GTK_WIDGET(window),"DownLoadFileTree");
  g_assert(view);
  /*
   *Do send
   */
  sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(view));  
  pathes=gtk_tree_selection_get_selected_rows(sel,&model);
  for(node=g_list_first(pathes);node;node=g_list_next(node)) {

    g_assert( (node) && (node->data));

    gtk_tree_model_get_iter(model,&iter,node->data);
    rc=do_download(model,node->data,&iter,window);
    if ( (rc < 0) && (!GTK_IS_WINDOW(window)) ){
      /*  window destroyed */
      g_list_foreach (pathes, gtk_tree_path_free, NULL);
      g_list_free (pathes);
      return;
    }
  }
  g_list_foreach (pathes, gtk_tree_path_free, NULL);
  g_list_free (pathes);

  /*
   * remove
   */
  for(sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
      gtk_tree_selection_count_selected_rows (sel)>0;){
      
    pathes=gtk_tree_selection_get_selected_rows(sel,&model);
    node=g_list_first(pathes);
    g_assert( (node) && (node->data));
    gtk_tree_model_get_iter(model,&iter,node->data);
    gtk_list_store_remove(GTK_LIST_STORE(model), &iter);    
    g_list_foreach (pathes, gtk_tree_path_free, NULL);
    g_list_free (pathes);
  }

  gtk_widget_set_sensitive(dirbtn,TRUE);

  noneed_to_destory = gtk_tree_model_get_iter_first (model, &iter);
  if (noneed_to_destory)
    dbg_out("some element remains\n");
  else
    gtk_widget_destroy (window);
}


void
on_DownLoadCancelBtn_clicked           (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *window;

  dbg_out("here:\n");
  window=lookup_widget(GTK_WIDGET(button),"downloadWindow");
  g_assert(window);
  gtk_widget_destroy (window);  
}


void
on_DownLoadOpenBtn_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
  int rc;
  GtkWidget *parent_window;
  GtkWidget *dialog;
  GtkWidget *file_entry;
  struct stat stat_buf;

  dbg_out("here\n");

  parent_window=lookup_widget(GTK_WIDGET(button),"downloadWindow");
  file_entry=lookup_widget(GTK_WIDGET(button),"DownLoadDirectoryEntry");
  /* 
   * なんで, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDERが
   *  アンドキュメンテッドなの?
   */
  dialog = gtk_file_chooser_dialog_new ("Download Directory",
					GTK_WINDOW(parent_window),
					GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					NULL);

  if (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
  {
    char *filename;

    filename = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (dialog));
    rc=stat(filename,&stat_buf);
    if (rc<0) 
      goto free_file_out;
    if (!(S_ISDIR(stat_buf.st_mode)))
      goto free_file_out;
    gtk_entry_set_text(GTK_ENTRY(file_entry), filename);
  free_file_out:
    g_free (filename);
  }
  gtk_widget_destroy (dialog);
}


void
on_messageSeparator_destroy            (GtkObject       *object,
                                        gpointer         user_data)
{

}


void
on_messageSeparator_drag_data_received (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{
  g_assert(widget);
  g_assert(data);

  dbg_out("here:data %s\n",(char *)data->data);  
  on_message_window_drag_data_received(widget,data);
}


void
on_textview1_drag_data_received        (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{
  g_assert(widget);
  g_assert(data);

  dbg_out("here:data %s\n",(char *)data->data);  
  on_message_window_drag_data_received(widget,data);
}


void
on_attachFileEditorMainFrame_drag_data_received
                                        (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{
  int rc;

  dbg_out("here\n");

  if (!data)
    return;

  rc=on_attach_window_drag_data_received(widget,data);

  if (rc<0)
    err_out("Can not handle drag data %s (%d)\n",strerror(-rc),-rc);

  return;
}


void
on_AttachFilePathEntry_drag_data_received
                                        (GtkWidget       *widget,
                                        GdkDragContext  *drag_context,
                                        gint             x,
                                        gint             y,
                                        GtkSelectionData *data,
                                        guint            info,
                                        guint            time,
                                        gpointer         user_data)
{

  GList *files,*node;
  GtkWidget *entry;

  dbg_out("here\n");

  if ( (!widget) || (!data) || (!(data->data)) )
    return;

  entry=lookup_widget(widget,"AttachFilePathEntry");
  g_assert(entry);

  files= gnome_vfs_uri_list_parse((gchar *)data->data);
  if (g_list_length(files) == 1) {
    /*  単一ファイルの場合は, ファイルパスをエントリに格納  */
    node=g_list_first(files);
    g_assert(node);
    dbg_out("draged file: %s\n",gnome_vfs_uri_get_path(node->data));
    gtk_entry_set_text(GTK_ENTRY(entry),gnome_vfs_uri_get_path(node->data));
  }else{
    dbg_out("droped multiple files.\n");
  }
  gnome_vfs_uri_list_free(files);

}


void
on_entry3_changed                      (GtkEditable     *editable,
                                        gpointer         user_data)
{
  GtkWidget *addr_entry;
  GtkWidget *configAddBcastBtn;
  gchar *addr;

  configAddBcastBtn=lookup_widget(GTK_WIDGET(editable),"configAddBcastBtn");
  g_assert(configAddBcastBtn);
  addr_entry=lookup_widget(GTK_WIDGET(editable),"entry3");
  g_assert(addr_entry);

  addr=(char *)gtk_entry_get_text(GTK_ENTRY(addr_entry));
  if ( (addr) && (*addr) ) 
    gtk_widget_set_sensitive(configAddBcastBtn,TRUE);
  else
    gtk_widget_set_sensitive(configAddBcastBtn,FALSE);

}
void
on_downloadMonitor_destroy             (GtkObject       *object,
                                        gpointer         user_data)
{
  dbg_out("here\n");

  gtk_widget_destroy(GTK_WIDGET(object));
}


void
on_downloadMonitor_show                (GtkWidget       *widget,
                                        gpointer         user_data)
{
  GtkWidget *window;  
  GtkWidget *view;
  GtkTreeSelection *sel;

  window=lookup_widget(widget,"downloadMonitor");
  view=lookup_widget(GTK_WIDGET(window),"treeview5");
  g_assert(view);
  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));

  update_download_view(window);
  update_download_monitor_button(sel);
}


void
on_deleteBtn_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *window;  
  GtkWidget *view;
  GtkTreeSelection *sel;
  GtkTreeIter  selected_row;
  GtkTreeModel *model;
  long pkt_no;

  dbg_out("here\n");

  window=lookup_widget(GTK_WIDGET(button),"downloadMonitor");
  view=lookup_widget(GTK_WIDGET(window),"treeview5");
  g_assert(view);
  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));

  if (gtk_tree_selection_get_selected(sel, &model, &selected_row)) {
    gtk_tree_model_get(model, &selected_row, 
		       3, &pkt_no,
		       -1);
    dbg_out("Relase:%ld\n",pkt_no);
    release_attach_file_block(pkt_no,TRUE);
    gtk_list_store_remove(GTK_LIST_STORE(model), &selected_row);
  }
}


void
on_updateBtn_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *window;  
  GtkWidget *view;
  GtkTreeSelection *sel;

  window=lookup_widget(GTK_WIDGET(button),"downloadMonitor");
  view=lookup_widget(GTK_WIDGET(window),"treeview5");
  g_assert(view);
  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));

  update_download_view(window);
  update_download_monitor_button(sel);
}


void
on_closeBtn_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *window;

  window=lookup_widget(GTK_WIDGET(button),"downloadMonitor");
  g_assert(window);
  gtk_widget_destroy(GTK_WIDGET(window));

}


void
on_initialWindow_check_resize          (GtkContainer    *container,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_downloadWindow_check_resize         (GtkContainer    *container,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_aboutdialog_check_resize            (GtkContainer    *container,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_viewWindow_check_resize             (GtkContainer    *container,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_attachFileEditor_check_resize       (GtkContainer    *container,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_downloadMonitor_check_resize        (GtkContainer    *container,
                                        gpointer         user_data)
{
  dbg_out("here\n");
}


void
on_initialWindow_size_allocate         (GtkWidget       *widget,
                                        GdkRectangle    *allocation,
                                        gpointer         user_data)
{
  g_assert(allocation);
  dbg_out("here:(%d,%d)\n",allocation->width,allocation->height);

}


void
on_downloadWindow_size_allocate        (GtkWidget       *widget,
                                        GdkRectangle    *allocation,
                                        gpointer         user_data)
{
  g_assert(allocation);
  dbg_out("here:(%d,%d)\n",allocation->width,allocation->height);
}


void
on_aboutdialog_size_allocate           (GtkWidget       *widget,
                                        GdkRectangle    *allocation,
                                        gpointer         user_data)
{
  g_assert(allocation);
  dbg_out("here:(%d,%d)\n",allocation->width,allocation->height);

}


void
on_viewWindow_size_allocate            (GtkWidget       *widget,
                                        GdkRectangle    *allocation,
                                        gpointer         user_data)
{
  g_assert(allocation);
  dbg_out("here:(%d,%d)\n",allocation->width,allocation->height);
  hostinfo_set_ipmsg_recv_window_size(allocation->width,
				      allocation->height);

}


void
on_attachFileEditor_size_allocate      (GtkWidget       *widget,
                                        GdkRectangle    *allocation,
                                        gpointer         user_data)
{
  g_assert(allocation);
  dbg_out("here:(%d,%d)\n",allocation->width,allocation->height);

}


void
on_downloadMonitor_size_allocate       (GtkWidget       *widget,
                                        GdkRectangle    *allocation,
                                        gpointer         user_data)
{
  g_assert(allocation);
  dbg_out("here:(%d,%d)\n",allocation->width,allocation->height);

}


void
on_messageWindow_size_allocate         (GtkWidget       *widget,
                                        GdkRectangle    *allocation,
                                        gpointer         user_data)
{
  g_assert(allocation);
  dbg_out("here:(%d,%d)\n",allocation->width,allocation->height);
  hostinfo_set_ipmsg_message_window_size(allocation->width,
					 allocation->height);
}



gboolean
on_textview1_button_press_event        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{

  return FALSE;
}

gboolean
on_viewwindowTextView_button_press_event
                                        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  GtkTextView *text_view;
  GtkTextIter start_iter,end_iter;
  GtkTextBuffer *buffer;
  gchar *string;
  gchar *url_string;
  gchar *url;

  dbg_out("here\n");

  switch (event->button) {
  case 1:
    text_view=lookup_widget(widget,"viewwindowTextView");
    g_assert(text_view);

    buffer=gtk_text_view_get_buffer(text_view);
    if (gtk_text_buffer_get_has_selection(buffer)) {

      dbg_out("selected\n");
      gtk_text_buffer_get_selection_bounds(buffer,
			     &start_iter,
			     &end_iter);
      string=gtk_text_buffer_get_text(buffer,
				  &start_iter,
				  &end_iter,
				  FALSE);
      dbg_out("String:%s\n",string);

      url_string=NULL;
      /* Shift JIS の ~がUTF-8の半角￣に置き換えられるので 
       * 一度SHIFT JISに戻す
       */
      convert_string_ipmsg_proto(string,&url_string);
      if (url_string) {
	  dbg_out("url-string:%s\n",url_string);
	  gnome_vfs_url_show (url_string);
	  g_free(url_string);
      }
      g_free(string);
    }
    
  default:					
    return FALSE;
    break;
  }
  return TRUE;
}

