#include <errno.h>
#include <glib.h>
#include "common.h"

#define IPMSG_INTERNAL_CODE "UTF-8"
#define IPMSG_PROTO_CODE    "SHIFT-JIS"
int
convert_string_internal(const char *string,const gchar **to_string){
  gchar *converted_string;
  gsize read_len;
  gsize write_len;
  GError *error_info=NULL;

  if ( (!string) || (!to_string) )
    return -EINVAL;

  converted_string=g_convert((const gchar *)string,
			     -1, /* ヌルターミネート */
			     IPMSG_INTERNAL_CODE,
			     IPMSG_PROTO_CODE,
			     &read_len,
			     &write_len,
			     &error_info);
  if (!converted_string) {
    return (error_info)?(error_info->code):(-EINVAL);
  }
  *to_string=converted_string;

  return 0;
}

int
convert_string_ipmsg_proto(const char *string,const gchar **to_string){
  gchar *converted_string;
  gsize read_len;
  gsize write_len;
  GError *error_info=NULL;

  if ( (!string) || (!to_string) )
    return -EINVAL;

  converted_string=g_convert((const gchar *)string,
			     -1, /* ヌルターミネート */
			     IPMSG_PROTO_CODE,
			     IPMSG_INTERNAL_CODE,
			     &read_len,
			     &write_len,
			     &error_info);
  if (!converted_string) {
    return (error_info)?(error_info->code):(-EINVAL);
  }
  *to_string=converted_string;

  return 0;
}

int
convert_string_ipmsg_locale(const char *string,const gchar **to_string){
  gchar *converted_string;
  gsize read_len;
  gsize write_len;
  GError *error_info=NULL;

  if ( (!string) || (!to_string) )
    return -EINVAL;

  converted_string=g_locale_from_utf8((const gchar *)string,
			     -1, /* ヌルターミネート */
			     &read_len,
			     &write_len,
			     &error_info);
  if (!converted_string) {
    err_out("%s\n",error_info->message);
    return (error_info)?(error_info->code):(-EINVAL);
  }
  *to_string=converted_string;

  return 0;
}

int
convert_string_ipmsg_filename(const char *string,const gchar **to_string){
  gchar *converted_string;
  gsize read_len;
  gsize write_len;
  GError *error_info=NULL;

  if ( (!string) || (!to_string) )
    return -EINVAL;

  converted_string=g_filename_from_utf8((const gchar *)string,
			     -1, /* ヌルターミネート */
			     &read_len,
			     &write_len,
			     &error_info);
  if (!converted_string) {
    err_out("%s\n",error_info->message);
    return (error_info)?(error_info->code):(-EINVAL);
  }
  *to_string=converted_string;

  return 0;
}
