#include <stdio.h>
#include <glib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <gconf/gconf-client.h>
#include "common.h"

static GConfClient *client=NULL;
static gchar *groupname=NULL;
static gchar *nickname=NULL;
static gchar *groupname_proto=NULL;
static gchar *nickname_proto=NULL;
static gchar *hostname=NULL;
static gchar *logfile_string=NULL;
static gchar *cite_string=NULL;
static gboolean current_absent_state=FALSE;
static int absent_id=-1;
GStaticMutex hostinfo_mutex = G_STATIC_MUTEX_INIT;

static const char *keys[] = {
  HOSTINFO_KEY_PORT,
  HOSTINFO_KEY_GROUP,
  HOSTINFO_KEY_NICKNAME,
  HOSTINFO_KEY_MSGSEC,
  HOSTINFO_KEY_CONFIRM_MSG,
  HOSTINFO_KEY_BROADCASTS,
  HOSTINFO_KEY_POPUP,
  HOSTINFO_KEY_SOUND,
  HOSTINFO_KEY_ENCLOSE,
  HOSTINFO_KEY_CITATION,
  HOSTINFO_KEY_LOGFILEPATH,
  HOSTINFO_KEY_ENABLE_LOG,
  HOSTINFO_KEY_LOG_NAME,
  HOSTINFO_KEY_LOG_IPADDR,
  HOSTINFO_KEY_CITE_STRING,
  HOSTINFO_KEY_ABS_TITLE,
  HOSTINFO_KEY_ABS_MSGS,
  HOSTINFO_KEY_DEBUG,
  HOSTINFO_KEY_MSGWIN_WIDTH,
  HOSTINFO_KEY_MSGWIN_HEIGHT,
  HOSTINFO_KEY_RECVWIN_WIDTH,
  HOSTINFO_KEY_RECVWIN_HEIGHT,
  NULL
};
gboolean
hostinfo_refer_debug_state(void) {
  if (!client)
    return FALSE;
  return gconf_client_get_bool(client,HOSTINFO_KEY_DEBUG,NULL);
}
int
hostinfo_send_broad_cast(const udp_con_t *con,const char *msg,size_t len){
  GSList *addr_top,*addr_list;

  udp_send_broadcast(con,msg,len);

  addr_top=addr_list=hostinfo_get_ipmsg_broadcast_list();
  while(addr_list != NULL) {
      gchar *element = addr_list->data;
      dbg_out("Send Broad cast:%s\n",element);
      udp_send_broadcast_with_addr(con,element,msg,len);
      g_free(element);
      addr_list=g_slist_next(addr_list);
  }
  if (addr_top)
    g_slist_free(addr_top);

  return 0;
}
int
hostinfo_set_ipmsg_broadcast_list(GSList *list){
  if (!list)
    return -EINVAL;
  
  gconf_client_set_list(client,HOSTINFO_KEY_BROADCASTS,GCONF_VALUE_STRING,list,NULL);
  gconf_client_clear_cache(client);
  return 0;
}
GSList*
hostinfo_get_ipmsg_broadcast_list(void){
  return gconf_client_get_list(client,HOSTINFO_KEY_BROADCASTS,GCONF_VALUE_STRING,NULL);
}
int 
hostinfo_refer_absent_length(int *max_index){
  int rc=-ESRCH;
  GSList *title_top,*title_list;

  if (!max_index)
    return -EINVAL;

  title_top=title_list=gconf_client_get_list(client,HOSTINFO_KEY_ABS_TITLE,GCONF_VALUE_STRING,NULL);

  if (title_top) {
    *max_index=g_slist_length(title_top);
    rc=0;
  }

  while(title_list != NULL) {
      gchar *element = title_list->data;
      dbg_out("free absent title:%s\n",element);
      g_free(element);
      title_list=g_slist_next(title_list);
  }
  if (title_top)
    g_slist_free(title_top);

  return rc;
}
int
hostinfo_get_absent_title(int index,const char **title){
  int rc=0;
  GSList *title_top,*title_list;
  GSList *title_ref;
  
  title_top=title_list=gconf_client_get_list(client,HOSTINFO_KEY_ABS_TITLE,GCONF_VALUE_STRING,NULL);
  if (!title_top)
    return -ESRCH;

  title_ref=g_slist_nth(title_top,index);
  if (title_ref) {
    *title=(title_ref->data)?(title_ref->data):(_("Undefined"));
    title_top=g_slist_remove_link(title_top,title_ref);
    g_slist_free_1(title_ref);
  }else{
    rc=-ESRCH;
  }
  title_list=title_top;
  while(title_list != NULL) {
      gchar *element = title_list->data;
      dbg_out("Free absent title:%s\n",element);
      g_free(element);
      title_list=g_slist_next(title_list);
  }
  if (title_top)
    g_slist_free(title_top);

  return rc;
}
int
hostinfo_set_ipmsg_absent_title(int index,const char *title){
  int rc=0;
  GSList *title_top,*title_list;
  GSList *title_ref;

  if (!title)
    return -EINVAL;

  title_top=title_list=gconf_client_get_list(client,HOSTINFO_KEY_ABS_TITLE,GCONF_VALUE_STRING,NULL);
  if (!title_top)
    return -ESRCH;

  rc=-ESRCH;
  title_ref=g_slist_nth(title_top,index);
  if (!title_ref) 
    goto free_list_out;

  g_free(title_ref->data);
  title_ref->data=strdup(title);
  if (!(title_ref->data)) {
    rc=-ENOMEM;
    title_top=g_slist_remove_link(title_top,title_ref);
    g_slist_free_1(title_ref);
    goto free_list_out;
  }

  gconf_client_set_list(client,HOSTINFO_KEY_ABS_TITLE,GCONF_VALUE_STRING,title_top,NULL);
  gconf_client_clear_cache(client);
  rc=0;

 free_list_out:
  title_list=title_top;
  while(title_list != NULL) {
      gchar *element = title_list->data;
      dbg_out("Free absent title:%s\n",element);
      g_free(element);
      title_list=g_slist_next(title_list);
  }
  if (title_top)
    g_slist_free(title_top);

  return rc;
}
/* absent msgs */
int 
hostinfo_refer_absent_message_slots(int *max_index){
  int rc=-ESRCH;
  GSList *message_top,*message_list;

  if (!max_index)
    return -EINVAL;

  message_top=message_list=gconf_client_get_list(client,HOSTINFO_KEY_ABS_MSGS,GCONF_VALUE_STRING,NULL);

  if (message_top) {
    *max_index=g_slist_length(message_top);
    rc=0;
  }

  while(message_list != NULL) {
      gchar *element = message_list->data;
      dbg_out("free absent message:%s\n",element);
      g_free(element);
      message_list=g_slist_next(message_list);
  }
  if (message_top)
    g_slist_free(message_top);

  return rc;
}
int
hostinfo_get_absent_message(int index,const char **message){
  int rc=0;
  GSList *message_top,*message_list;
  GSList *message_ref;
  
  message_top=message_list=gconf_client_get_list(client,HOSTINFO_KEY_ABS_MSGS,GCONF_VALUE_STRING,NULL);
  if (!message_top)
    return -ESRCH;

  message_ref=g_slist_nth(message_top,index);
  if (message_ref) {
    *message=(message_ref->data)?(message_ref->data):(_("Undefined"));
    message_top=g_slist_remove_link(message_top,message_ref);
    g_slist_free_1(message_ref);
  }else{
    rc=-ESRCH;
  }
  message_list=message_top;
  while(message_list != NULL) {
      gchar *element = message_list->data;
      dbg_out("Free absent message:%s\n",element);
      g_free(element);
      message_list=g_slist_next(message_list);
  }
  if (message_top)
    g_slist_free(message_top);

  return rc;
}
int
hostinfo_set_ipmsg_absent_message(int index,const char *message){
  int rc=0;
  GSList *message_top,*message_list;
  GSList *message_ref;

  if (!message)
    return -EINVAL;

  message_top=message_list=gconf_client_get_list(client,HOSTINFO_KEY_ABS_MSGS,GCONF_VALUE_STRING,NULL);
  if (!message_top)
    return -ESRCH;

  rc=-ESRCH;
  message_ref=g_slist_nth(message_top,index);
  if (!message_ref) 
    goto free_list_out;

  g_free(message_ref->data);
  message_ref->data=strdup(message);
  if (!(message_ref->data)) {
    rc=-ENOMEM;
    message_top=g_slist_remove_link(message_top,message_ref);
    g_slist_free_1(message_ref);
    goto free_list_out;
  }

  gconf_client_set_list(client,HOSTINFO_KEY_ABS_MSGS,GCONF_VALUE_STRING,message_top,NULL);
  gconf_client_clear_cache(client);
  rc=0;

 free_list_out:
  message_list=message_top;
  while(message_list != NULL) {
      gchar *element = message_list->data;
      dbg_out("Free absent message:%s\n",element);
      g_free(element);
      message_list=g_slist_next(message_list);
  }
  if (message_top)
    g_slist_free(message_top);

  return rc;
}
/* ids */
int 
hostinfo_get_absent_id(int *index){
  int max_index;
  int max_slots;

  if (!index)
    return -EINVAL;
  if (hostinfo_refer_absent_length(&max_index))
    return -EINVAL;
  if (hostinfo_refer_absent_message_slots(&max_slots))
    return -EINVAL;
  g_assert(max_index==max_slots);
  if (absent_id<0)
    return -EINVAL;
  if (absent_id>max_index)
    return -EINVAL;

  *index=absent_id;
  return 0;
}
int 
hostinfo_set_absent_id(int index){
  int max_index;
  int max_slots;

  if (index<0)
    return -EINVAL;
  if (hostinfo_refer_absent_length(&max_index))
    return -EINVAL;
  if (hostinfo_refer_absent_message_slots(&max_slots))
    return -EINVAL;
  g_assert(max_index==max_slots);

  if (index>max_index)
    return -EINVAL;
  absent_id=index;
  return 0;
}

gboolean
hostinfo_is_ipmsg_absent(void) {

  return current_absent_state;
}
gboolean
hostinfo_set_ipmsg_absent(gboolean state) {
  current_absent_state=state;
  return current_absent_state;
}

gboolean
hostinfo_refer_ipmsg_default_secret(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_MSGSEC,NULL);
}
gboolean
hostinfo_set_ipmsg_default_secret(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_MSGSEC,val,NULL);
}
gboolean
hostinfo_refer_ipmsg_default_confirm(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_CONFIRM_MSG,NULL);
}
gboolean
hostinfo_set_ipmsg_default_confirm(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_CONFIRM_MSG,val,NULL);
}
gboolean
hostinfo_refer_ipmsg_default_popup(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_POPUP,NULL);
}
gboolean
hostinfo_set_ipmsg_default_popup(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_POPUP,val,NULL);
}

gboolean
hostinfo_refer_ipmsg_default_sound(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_SOUND,NULL);
}
gboolean
hostinfo_set_ipmsg_default_sound(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_SOUND,val,NULL);
}
gboolean
hostinfo_refer_ipmsg_default_enclose(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_ENCLOSE,NULL);
}
gboolean
hostinfo_set_ipmsg_default_enclose(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_ENCLOSE,val,NULL);
}

gboolean
hostinfo_refer_ipmsg_default_citation(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_CITATION,NULL);
}
gboolean
hostinfo_set_ipmsg_default_citation(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_CITATION,val,NULL);
}

gboolean
hostinfo_refer_ipmsg_ipaddr_logging(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_LOG_IPADDR,NULL);
}
gboolean
hostinfo_set_ipmsg_ipaddr_logging(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_LOG_IPADDR,val,NULL);
}

gboolean
hostinfo_refer_ipmsg_logname_logging(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_LOG_NAME,NULL);
}
gboolean
hostinfo_set_ipmsg_logname_logging(gboolean val) {

  return gconf_client_set_bool(client,HOSTINFO_KEY_LOG_NAME,val,NULL);
}

gboolean
hostinfo_refer_ipmsg_enable_log(void) {

  return gconf_client_get_bool(client,HOSTINFO_KEY_ENABLE_LOG,NULL);
}
gboolean
hostinfo_set_ipmsg_enable_log(gboolean val) {


  return gconf_client_set_bool(client,HOSTINFO_KEY_ENABLE_LOG,val,NULL);
}
int 
hostinfo_set_ipmsg_logfile(const char *file) {
  gchar *string;
  int rc;

  if (!file)
    return -EINVAL;

  rc=logfile_reopen_logfile(file);
  if (rc)
    return rc;

  g_static_mutex_lock(&hostinfo_mutex);  

  if (logfile_string)
    g_free(logfile_string);

  logfile_string=NULL;

  gconf_client_set_string(client,HOSTINFO_KEY_LOGFILEPATH,file,NULL);

  string=gconf_client_get_string(client,HOSTINFO_KEY_LOGFILEPATH,NULL);
  if (string) {
    dbg_out("gconf return logfile %s\n",string);
    g_static_mutex_unlock(&hostinfo_mutex);  
    logfile_string=string;
    return 0;
  }
  g_static_mutex_unlock(&hostinfo_mutex);  
  return -ENOMEM;
}
const char *
hostinfo_refer_ipmsg_logfile(void) {
  char *string;

  g_static_mutex_lock(&hostinfo_mutex);  
  if (logfile_string)
    goto unlock_out;

  string=gconf_client_get_string(client,HOSTINFO_KEY_LOGFILEPATH,NULL);

  if (string) {
    dbg_out("gconf return logfile %s\n",string);
    g_static_mutex_unlock(&hostinfo_mutex);  
    logfile_string=string;
  }  
 unlock_out:
  g_static_mutex_unlock(&hostinfo_mutex);  
  return (logfile_string)?(logfile_string):("");
}

const char *
hostinfo_refer_ipmsg_cite_string(void) {
  char *string;

  if (cite_string)
    return cite_string;

  g_static_mutex_lock(&hostinfo_mutex);
  string=gconf_client_get_string(client,HOSTINFO_KEY_CITE_STRING,NULL);

  if (string) {
    dbg_out("gconf return cite string:%s\n",string);
    cite_string=string;
    g_static_mutex_unlock(&hostinfo_mutex);  
    return (const char *)cite_string;
  }
  g_static_mutex_unlock(&hostinfo_mutex);  
  return ">";
}
int
hostinfo_refer_ipmsg_port(void){
  int port;

  g_static_mutex_lock(&hostinfo_mutex);  
  port=gconf_client_get_int(client,HOSTINFO_KEY_PORT,NULL);
  dbg_out("gconf return port:%d\n",port);
  g_static_mutex_unlock(&hostinfo_mutex);  
  if (!port) 
    return DEFAULT_PORT;
  
  return port;
}
int
hostinfo_set_ipmsg_port(int port){
  int new;

  g_static_mutex_lock(&hostinfo_mutex);  
  gconf_client_set_int(client,HOSTINFO_KEY_PORT,port,NULL);
  gconf_client_clear_cache(client);
  new=gconf_client_get_int(client,HOSTINFO_KEY_PORT,NULL);
  dbg_out("gconf return port:%d\n",new);
  g_static_mutex_unlock(&hostinfo_mutex);  
  if (!new)
    return DEFAULT_PORT;

  return port;
}

const char *
hostinfo_refer_user_name(void) {
  struct passwd *ref_p;

  g_static_mutex_lock(&hostinfo_mutex);  
  ref_p=getpwuid(getuid());

  if (!ref_p) {
    err_out("Can not get pwent:%s (%d)\n",strerror(errno),errno);
    g_static_mutex_unlock(&hostinfo_mutex);  
    return _("UnKnown");
  }
  dbg_out("UserName:%s\n",ref_p->pw_name);
  g_static_mutex_unlock(&hostinfo_mutex);  
  return ref_p->pw_name;
}

const char *
hostinfo_refer_group_name(void) {

  g_static_mutex_lock(&hostinfo_mutex);  
  if (!groupname) {
    groupname=gconf_client_get_string(client,HOSTINFO_KEY_GROUP,NULL);
    dbg_out("gconf return groupname:%s\n",groupname);
  }
  g_static_mutex_unlock(&hostinfo_mutex);  
  return (groupname)?(groupname):(_("Default Group"));
}
int
hostinfo_set_group_name(const char *groupName) {

  if (!groupName)
    return -EINVAL;

  dbg_out("set groupname:%s\n",groupName);

  g_static_mutex_lock(&hostinfo_mutex);  
  if (groupname)
    g_free(groupname);

  if (groupname_proto)
    g_free(groupname_proto);

  g_assert(client);

  gconf_client_set_string(client,HOSTINFO_KEY_GROUP,groupName,NULL);
  gconf_client_clear_cache(client);
  groupname=gconf_client_get_string(client,HOSTINFO_KEY_GROUP,NULL);
  dbg_out("gconf return groupname:%s\n",groupname);
  convert_string_ipmsg_proto(groupname,(const gchar **)&groupname_proto);
  g_static_mutex_unlock(&hostinfo_mutex);  
  return 0;
}

int
hostinfo_set_nick_name(const char *nickName) {

  if (!nickName)
    return -EINVAL;

  g_static_mutex_lock(&hostinfo_mutex);  
  dbg_out("set nickName:%s\n",nickName);

  g_assert(client);

  if (nickname)
    g_free(nickname);
  if (nickname_proto)
    g_free(nickname_proto);

  gconf_client_set_string(client,HOSTINFO_KEY_NICKNAME,nickName,NULL);
  gconf_client_clear_cache(client);
  nickname=gconf_client_get_string(client,HOSTINFO_KEY_NICKNAME,NULL);
  dbg_out("gconf return nick name:%s\n",nickname);
  convert_string_ipmsg_proto(nickname,(const gchar **)&nickname_proto);
  g_static_mutex_unlock(&hostinfo_mutex);  
  return 0;
}

const char *
hostinfo_refer_group_name_proto(void) {
  char *internal_name;

  g_static_mutex_lock(&hostinfo_mutex);  
  if (!groupname_proto) {
    internal_name=gconf_client_get_string(client,HOSTINFO_KEY_GROUP,NULL);
    convert_string_ipmsg_proto(internal_name,(const gchar **)&groupname_proto);
    dbg_out("gconf return groupname:%s\n",groupname_proto);
    g_free(internal_name);
  }
  g_static_mutex_unlock(&hostinfo_mutex);  

  return groupname_proto;
}

const char *
hostinfo_refer_nick_name(void) {

  g_static_mutex_lock(&hostinfo_mutex);  
  if (!nickname) {
    nickname=gconf_client_get_string(client,HOSTINFO_KEY_NICKNAME,NULL);
    dbg_out("gconf return nick name:%s\n",nickname);
    if (!nickname_proto)
      convert_string_ipmsg_proto(nickname,(const gchar **)&nickname_proto);
  }
  g_static_mutex_unlock(&hostinfo_mutex);  
  return (nickname)?(nickname):(_("Default Nick Name"));
}

const char *
hostinfo_refer_nick_name_proto(void) {
  gchar *internal_name;

  g_static_mutex_lock(&hostinfo_mutex);  
  if (!nickname_proto) {
    internal_name=gconf_client_get_string(client,HOSTINFO_KEY_NICKNAME,NULL);
    convert_string_ipmsg_proto(internal_name,(const gchar **)&nickname_proto);
    g_free(internal_name);
    dbg_out("gconf return nickname:%s\n",nickname_proto);
  }
  g_static_mutex_unlock(&hostinfo_mutex);  
  return nickname_proto;
}

const char *
hostinfo_refer_host_name(void){

  if (hostname)
    goto ret;

  g_static_mutex_lock(&hostinfo_mutex);  
  hostname=malloc(64);

  if (gethostname(hostname,64) < 0) {
    free(hostname);
    hostname=NULL;
    g_static_mutex_unlock(&hostinfo_mutex);  
    return _("Unknown");
  }
  hostname[63]=0;
  g_static_mutex_unlock(&hostinfo_mutex);  
 ret:
  return hostname;
}


static void
hostinfo_gconf_notify_func(GConfClient *client, guint cnxn_id, GConfEntry  *entry,
			   gpointer user_data){
  dbg_out("Configuration is changed: key=%s\n",(char *)user_data);
}
int 
hostinfo_get_normal_send_flags(void){
  int flags=0;

  g_static_mutex_lock(&hostinfo_mutex);  
  flags=IPMSG_SENDCHECKOPT; /* 受信確認をデフォルト  */

  if (gconf_client_get_bool(client,HOSTINFO_KEY_MSGSEC,NULL))
    flags|=IPMSG_SECRETOPT;

  if (gconf_client_get_bool(client,HOSTINFO_KEY_CONFIRM_MSG,NULL))
    flags|=IPMSG_SENDCHECKOPT;

  dbg_out("gconf return flags:%x\n",flags);
  g_static_mutex_unlock(&hostinfo_mutex);  
  return flags;
}
static int 
hostinfo_get_ipmsg_window_size(const gchar *wkey,const gchar *hkey,gint *width,gint *height){

  dbg_out("here\n");

  if ( (!wkey) || (!hkey) ||  (!width) || (!height) )
    return -EINVAL;

  g_static_mutex_lock(&hostinfo_mutex);
  *width=gconf_client_get_int(client,wkey,NULL);
  *height=gconf_client_get_int(client,hkey,NULL);
  g_static_mutex_unlock(&hostinfo_mutex);

  dbg_out("gconf return size:(%s,%s)=(%d,%d)\n",wkey,hkey,*width,*height);

  return 0;
}
static int 
hostinfo_set_ipmsg_window_size(const gchar *wkey,const gchar *hkey,gint width,gint height){
  int rwidth,rheight;

  dbg_out("here\n");

  if ( (!wkey) || (!hkey) )
    return -EINVAL;

  dbg_out("gconf set size:(%s,%s)=(%d,%d)\n",wkey,hkey,width,height);


  g_static_mutex_lock(&hostinfo_mutex);

  gconf_client_set_int(client,wkey,width,NULL);
  gconf_client_set_int(client,hkey,height,NULL);

  gconf_client_clear_cache(client);

  rwidth=gconf_client_get_int(client,wkey,NULL);
  rheight=gconf_client_get_int(client,hkey,NULL);

  g_static_mutex_unlock(&hostinfo_mutex);


  dbg_out("gconf get size:(%s,%s)=(%d,%d)\n",wkey,hkey,rwidth,rheight);

  return 0;
}

int 
hostinfo_get_ipmsg_message_window_size(gint *width,gint *height){
  int rc;
  if ( (!width) || (!height) )
    return -EINVAL;

  rc=hostinfo_get_ipmsg_window_size(HOSTINFO_KEY_MSGWIN_WIDTH,
				 HOSTINFO_KEY_MSGWIN_HEIGHT,
				 width,
				 height);
  dbg_out("gconf return size:(%d,%d)\n",*width,*height);

  return rc;
}
int 
hostinfo_set_ipmsg_message_window_size(gint width,gint height){
  int rc;

  dbg_out("gconf set size:(%d,%d)\n",width,height);
  rc=hostinfo_set_ipmsg_window_size(HOSTINFO_KEY_MSGWIN_WIDTH,
				 HOSTINFO_KEY_MSGWIN_HEIGHT,
				 width,
				 height);
  return rc;
}
int 
hostinfo_get_ipmsg_recv_window_size(gint *width,gint *height){
  int rc;
  if ( (!width) || (!height) )
    return -EINVAL;

  rc=hostinfo_get_ipmsg_window_size(HOSTINFO_KEY_RECVWIN_WIDTH,
				 HOSTINFO_KEY_RECVWIN_HEIGHT,
				 width,
				 height);
  dbg_out("gconf return size:(%d,%d)\n",*width,*height);

  return rc;
}
int 
hostinfo_set_ipmsg_recv_window_size(gint width,gint height){
  int rc;

  dbg_out("gconf set size:(%d,%d)\n",width,height);
  rc=hostinfo_set_ipmsg_window_size(HOSTINFO_KEY_RECVWIN_WIDTH,
				 HOSTINFO_KEY_RECVWIN_HEIGHT,
				 width,
				 height);
  return rc;
}
int
hostinfo_init_hostinfo(void){
  int i;

  client = gconf_client_get_default ();

  gconf_client_add_dir (client,
			PATH,
			GCONF_CLIENT_PRELOAD_NONE,
			NULL);
  /*
   * Add keys
   */
  for(i=0;keys[i];++i) {
    dbg_out("Add %s\n",keys[i]);
    gconf_client_notify_add (client,
			     keys[i],
			     hostinfo_gconf_notify_func,
			     (gpointer)keys[i],
			     NULL,
			     NULL);
  }
  return 0;
}
void
hostinfo_cleanup_hostinfo(void){
  if (groupname)
    g_free(groupname);
  if (nickname)
    g_free(nickname);
  if (groupname_proto)
    g_free(groupname_proto);
  if (nickname_proto)
    g_free(nickname_proto);
  if (hostname)
    g_free(hostname);
  if (cite_string)
    g_free(cite_string);
}
