#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <bonobo.h>
#include <gnome.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "common.h"

#define GLADE_HOOKUP_OBJECT_NO_REF(component,widget,name) \
  g_object_set_data (G_OBJECT (component), name, widget)

#define GLADE_HOOKUP_OBJECT(component,widget,name) \
  g_object_set_data_full (G_OBJECT (component), name, \
    gtk_widget_ref (widget), (GDestroyNotify) gtk_widget_unref)

static GtkWidget *
create_menu_item(const char *name,const char *label,gpointer data,
		 void (*menu_item_callback_function)
		 (gpointer user_data)){
  GtkWidget *new_item;
  int index=(int)data;

  dbg_out("Create item with:%d\n",index);
  new_item=gtk_menu_item_new_with_label(label);
  if (menu_item_callback_function){
    dbg_out("connect:data %x\n",(unsigned int)data);
    if (data)
      gtk_signal_connect_object( (gpointer)new_item, "activate",
				 GTK_SIGNAL_FUNC(menu_item_callback_function), 
				 data);
    else
      gtk_signal_connect_object( (gpointer)new_item, "activate",
				 GTK_SIGNAL_FUNC(menu_item_callback_function), 
				 (gpointer )new_item);
  }
  gtk_widget_show(new_item);

  return new_item;
}
static GtkWidget *
create_stock_menu_item(const char *name,const gchar *stock_id,gpointer data,
		 void (*menu_item_callback_function)
		 (gpointer user_data)){
  GtkWidget *new_item;

  new_item=gtk_image_menu_item_new_from_stock(stock_id,NULL);
  if (menu_item_callback_function) {
    if (data)
      gtk_signal_connect_object( GTK_OBJECT(new_item), "activate",
				 GTK_SIGNAL_FUNC(menu_item_callback_function), 
				 data);
    else
      gtk_signal_connect_object( GTK_OBJECT(new_item), "activate",
				 GTK_SIGNAL_FUNC(menu_item_callback_function), 
				 new_item);
  }
  gtk_widget_show(new_item);

  return new_item;
}
static GtkWidget *
create_menu_separator(const char *name){
  GtkWidget *new_item;

  new_item=gtk_separator_menu_item_new ();
  gtk_widget_show(new_item);

  return new_item;
}
void
update_download_monitor_button(GtkTreeSelection *sel) {
  GtkTreeIter  selected_row;
  GtkWidget    *view;
  GtkWidget    *remove_btn;

  view=GTK_WIDGET(gtk_tree_selection_get_tree_view(sel));
  g_assert(view);
  remove_btn=lookup_widget(GTK_WIDGET(view),"deleteBtn");
  g_assert(remove_btn);
  g_assert(gtk_tree_selection_get_mode(sel) == GTK_SELECTION_SINGLE);
  
  if (gtk_tree_selection_get_selected(sel, NULL, &selected_row))
    {
      gtk_widget_set_sensitive(remove_btn, TRUE);
    }
  else
    {
      gtk_widget_set_sensitive(remove_btn, FALSE);
    }
}
static void
onDownloadMonitorSelectionChanged (GtkTreeSelection *sel, GtkListStore *liststore) {
  update_download_monitor_button(sel);
}

void
setup_download_window(GtkWidget *window){
  GtkWidget *view;
  GtkListStore        *liststore;
  GtkTreeViewColumn   *col;
  GtkCellRenderer     *renderer;
  GtkTreeSelection    *sel;

  dbg_out("here\n");
  g_assert(window);
  view=lookup_widget(GTK_WIDGET(window),"treeview5");
  g_assert(view);
  dbg_out("Create new monitor\n");

  liststore = gtk_list_store_new(4, 
				 G_TYPE_STRING,
				 G_TYPE_UINT,
				 G_TYPE_STRING,
				 G_TYPE_INT);
  g_assert(liststore);
  gtk_tree_view_set_model(GTK_TREE_VIEW(view), GTK_TREE_MODEL(liststore));

    /*
   * タイトル設定
   */
  /* --- Column #1 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("FileNames"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 0);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #2 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("Remains"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 1);

  gtk_tree_view_column_set_resizable (col,TRUE);

  /* --- Column #3 --- */

  col = gtk_tree_view_column_new();

  gtk_tree_view_column_set_title(col, _("Receivers"));

  /* pack tree view column into tree view */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), col);

  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */
  gtk_tree_view_column_pack_start(col, renderer, TRUE);

  /* connect 'text' property of the cell renderer to
   *  model column that contains the first name */
  gtk_tree_view_column_add_attribute(col, renderer, "text", 2);

  gtk_tree_view_column_set_resizable (col,TRUE);

  sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
  gtk_tree_selection_set_mode(sel, GTK_SELECTION_SINGLE);
  g_signal_connect(sel, "changed", 
		   G_CALLBACK(onDownloadMonitorSelectionChanged),
		   liststore);
}
static void
release_download_info(gpointer data, gpointer user_data){
  download_monitor_entry_t *download_info;

  dbg_out("here\n");
  if (!data)
    return;

  download_info=data;
  dbg_out("Release:%ld\n",download_info->pkt_no);
  release_download_monitor_entry(download_info);
  
}
void 
update_download_view(GtkWidget *window) {
  GList *list=NULL;
  GList *node;
  GtkWidget *view;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;
  download_monitor_entry_t *download_info;

  dbg_out("here\n");

  g_assert(window);
  view=lookup_widget(GTK_WIDGET(window),"treeview5");
  g_assert(view);

  model = gtk_tree_view_get_model(GTK_TREE_VIEW(view));
  if (gtk_tree_model_get_iter_first(model,&iter)) {
    for(valid=gtk_tree_model_get_iter_first(model,&iter);
	valid;
	valid=gtk_tree_model_iter_next(model,&iter)) {
      gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
    }
  }

  list=get_download_monitor_info();
  for(node=g_list_first (list);
      (node);
      node=g_list_next(node)) {

    if (!(list->data))
      continue;
    
    download_info=(download_monitor_entry_t *)(node->data);
    gtk_list_store_append(GTK_LIST_STORE(model), &iter);
    gtk_list_store_set(GTK_LIST_STORE(model),
		       &iter,
		       0,download_info->filenames,
		       1,download_info->status,
		       2,download_info->username,
		       3,download_info->pkt_no,
		       -1);
  } 
  g_list_foreach(list,release_download_info,NULL);
  g_list_free(list);
}
void
on_mainmenu_download_item(gpointer menuitem){
  GtkWidget *window;

  dbg_out("here\n");
  
  window=create_downloadMonitor();
  g_assert(window);
  setup_download_window(window);
  gtk_widget_show(window);
}
void
on_mainmenu_preferences_item(gpointer menuitem){
  dbg_out("here\n");
  on_preferences1_activate(menuitem,NULL);
}
void
on_mainmenu_about_item(gpointer menuitem){
  dbg_out("here\n");

  ipmsg_show_about_dialog();
}
void
on_mainmenu_log_item(gpointer menuitem){
  dbg_out("here\n");
}
void
on_mainmenu_absent_item(gpointer menuitem){
  dbg_out("here\n");
}
void
on_mainmenu_attend_item(gpointer menuitem){
  dbg_out("here\n");
  hostinfo_set_ipmsg_absent(FALSE);
  ipmsg_send_br_absence(udp_con,0);
}
void
on_mainmenu_foreground_win_item(gpointer menuitem){
  dbg_out("here\n");
  present_all_displayed_windows();
}
void
on_mainmenu_remove_win_item(gpointer menuitem){
  dbg_out("here\n");
  destroy_all_opened_windows();
}
void
on_mainmenu_quit_item (gpointer menuitem){
  dbg_out("here\n");
  on_initialWindow_destroy(GTK_OBJECT(menuitem),NULL);
}
void
on_fuzai_item_activate (gpointer user_data){
  int max_index;
  int index;

  index=(int)user_data - 1;
  dbg_out("here:%x\n",index);
  if (hostinfo_refer_absent_length(&max_index)) {
    g_assert_not_reached();
  }
  dbg_out("Max index:%d\n",max_index);
  if ( (index >= max_index) || (index < 0) )
    return;
  hostinfo_set_absent_id(index);
  hostinfo_set_ipmsg_absent(TRUE);
  ipmsg_send_br_absence(udp_con,0);
  return;
}
void
on_fuzai_config_activate (gpointer menuitem){
  dbg_out("here:%x\n",(unsigned int)menuitem);
}

static GtkWidget *
create_fuzai_menu(void){
  GtkWidget *menu;
  GtkWidget *new_item;
  GtkWidget *separator4_item;
  GtkWidget *config_item;
  gchar name[16];
  gchar *title;
  int i,max_index=0;

  menu=gtk_menu_new(); 
  if (hostinfo_refer_absent_length(&max_index)) {
    g_assert_not_reached();
  }
  dbg_out("Max index:%d\n",max_index);
  for(i=0;i<max_index;++i) {
    snprintf(name,15,"fuzai%d",i);
    name[15]='\0';
    title=NULL;
    hostinfo_get_absent_title(i,(const char **)&title);
    /*  埋め込みのindexは, NULLと区別するため, +1する */
    if (title) {
      new_item=create_menu_item(name,title,(gpointer)(i+1),on_fuzai_item_activate);
      g_free(title);
    }else
      new_item=create_menu_item(name,name,(gpointer)(i+1),on_fuzai_item_activate);
    gtk_menu_append( GTK_MENU(menu), new_item);
    GLADE_HOOKUP_OBJECT(menu,new_item,name);
  }
  separator4_item=create_menu_separator("separator4");

  //  config_item=create_stock_menu_item("fuzai_config_item",GTK_STOCK_PREFERENCES,NULL,on_fuzai_config_activate);

  gtk_menu_append( GTK_MENU(menu), separator4_item);
  //  gtk_menu_append( GTK_MENU(menu), config_item);

  GLADE_HOOKUP_OBJECT(menu,separator4_item,"separator4_item");
  //  GLADE_HOOKUP_OBJECT(menu,config_item,"fuzai_config_item");

  return menu;
}
GtkWidget *
create_main_menu(void){
  GtkWidget *menu;
  GtkWidget *download_item;
  GtkWidget *separator1_item;
  GtkWidget *remove_win_item;
  GtkWidget *foreground_win_item;
  GtkWidget *separator2_item;
  GtkWidget *preferences_item;
  GtkWidget *about_item;
  GtkWidget *log_item;
  GtkWidget *separator3_item;
  GtkWidget *absent_item;
  GtkWidget *attend_item;
  GtkWidget *separator4_item;
  GtkWidget *quit_item;

  menu=gtk_menu_new(); 

  download_item=create_menu_item("download_item",_("DownLoadMonitor"),NULL,on_mainmenu_download_item);
  separator1_item=create_menu_separator("separator1");
  foreground_win_item=create_menu_item("foreground_win_item",_("Foreground all message windows"),NULL,on_mainmenu_foreground_win_item);
  remove_win_item=create_menu_item("remove_win_item",_("Remove opend windows"),NULL,on_mainmenu_remove_win_item);
  separator2_item=create_menu_separator("separator2");
  preferences_item=create_stock_menu_item("preferences_item",GTK_STOCK_PREFERENCES,NULL,on_mainmenu_preferences_item);
  about_item=create_stock_menu_item("about_item",GTK_STOCK_ABOUT,NULL,on_mainmenu_about_item);
  //  log_item=create_menu_item("log_item",_("Show Log"),NULL,on_mainmenu_log_item);
  separator3_item=create_menu_separator("separator3");
  absent_item=create_menu_item("absent_item",_("Absence"),NULL,on_mainmenu_absent_item);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (absent_item), create_fuzai_menu());

  if (hostinfo_is_ipmsg_absent())
    attend_item=create_menu_item("attend_item",_("Attend"),NULL,on_mainmenu_attend_item);

  separator4_item=create_menu_separator("separator4");
  quit_item=create_stock_menu_item("quit_item",GTK_STOCK_QUIT,NULL,on_mainmenu_quit_item);

  gtk_menu_append( GTK_MENU(menu), download_item);
  gtk_menu_append( GTK_MENU(menu), separator1_item);
  gtk_menu_append( GTK_MENU(menu), remove_win_item);
  gtk_menu_append( GTK_MENU(menu), foreground_win_item);
  gtk_menu_append( GTK_MENU(menu), separator2_item);
  gtk_menu_append( GTK_MENU(menu), preferences_item);
  gtk_menu_append( GTK_MENU(menu), about_item);
  //gtk_menu_append( GTK_MENU(menu), log_item);
  gtk_menu_append( GTK_MENU(menu), separator3_item);

  gtk_menu_append( GTK_MENU(menu), absent_item);
  if (hostinfo_is_ipmsg_absent())
    gtk_menu_append( GTK_MENU(menu), attend_item);

  gtk_menu_append( GTK_MENU(menu), separator4_item);

  gtk_menu_append( GTK_MENU(menu), quit_item);

  GLADE_HOOKUP_OBJECT(menu,download_item,"download_item");
  GLADE_HOOKUP_OBJECT(menu,separator1_item,"separator1_item");
  GLADE_HOOKUP_OBJECT(menu,remove_win_item,"remove_win_item");
  GLADE_HOOKUP_OBJECT(menu,foreground_win_item,"foreground_win_item");
  GLADE_HOOKUP_OBJECT(menu,separator1_item,"separator2_item");
  GLADE_HOOKUP_OBJECT(menu,preferences_item,"preferences_item");
  GLADE_HOOKUP_OBJECT(menu,about_item,"about_item");
  //GLADE_HOOKUP_OBJECT(menu,log_item,"log_item");
  GLADE_HOOKUP_OBJECT(menu,separator3_item,"separator3_item");
  
  GLADE_HOOKUP_OBJECT(menu,absent_item,"absent_item");
  if (hostinfo_is_ipmsg_absent())
    GLADE_HOOKUP_OBJECT(menu,attend_item,"attend_item");
  GLADE_HOOKUP_OBJECT(menu, separator4_item," separator4_item");

  GLADE_HOOKUP_OBJECT(menu,quit_item,"quit_item");

  return menu;
}
gboolean
on_init_win_event_button_press_event (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
	GtkWidget	*fuzaiMenu;

	dbg_out("buttonWin: button press: %d\n", event->button);
	switch (event->button) {
	case 3:
	  gtk_menu_popup(GTK_MENU(create_main_menu()), NULL, NULL, NULL, NULL,
			       event->button, event->time);

	default:					/* main menu */
	  return FALSE;
	  break;
	}
	return TRUE;
}

void
ipmsg_show_about_dialog(void){
  GdkPixbuf *aboutDialog_icon_pixbuf;
  GtkWidget *aboutDialog;
  const gchar *authors[]={
    G2IPMSG_AUTHOR,
    NULL
  };

  dbg_out("here\n");

  aboutDialog=create_aboutdialog ();
  g_assert(aboutDialog);

  gtk_about_dialog_set_authors(GTK_ABOUT_DIALOG(aboutDialog),(const gchar **)authors);
  gtk_about_dialog_set_translator_credits (GTK_ABOUT_DIALOG(aboutDialog),G2IPMSG_TRANSLATOR);
  gtk_about_dialog_set_copyright (GTK_ABOUT_DIALOG(aboutDialog),_("Takeharu KATO"));
  aboutDialog_icon_pixbuf = create_pixbuf ("g2ipmsg/ipmsg.xpm");
  if (aboutDialog_icon_pixbuf)
    {
      gtk_window_set_icon (GTK_WINDOW (aboutDialog), aboutDialog_icon_pixbuf);
      gdk_pixbuf_unref (aboutDialog_icon_pixbuf);
    }

  gtk_widget_show (aboutDialog);

}

