#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <glib.h>
#include <unistd.h>

#include "common.h"
static int
internal_set_buffer(const int soc,int type,int max,int min,int *actual){
  int rc;
  int size;

  if ( (soc<0) || (!actual) )
    return -EINVAL;
  if ( (type != SO_RCVBUF) && (type != SO_SNDBUF) )
    return -EINVAL;

  size=max;
  rc=setsockopt(soc, SOL_SOCKET, type, (void*)&size, sizeof(int));
  if (rc<0) {
    size=min;
    rc=setsockopt(soc, SOL_SOCKET, type, (void*)&size, sizeof(int));
    if (rc<0) {
      err_out("Can not set broad cast:%s(%d)\n",strerror(errno),errno);
      *actual=size;
      return -errno;
    }
  }
  *actual=size;
  return 0;
}
int
sock_set_buffer(int soc){
  int rc;
  int size=0;

  if (soc<0)
    return -EINVAL;

  rc=internal_set_buffer(soc,SO_RCVBUF,_MSG_BUF_SIZE,_MSG_BUF_MIN_SIZE,&size);
  if (size < _MSG_BUF_MIN_SIZE) {
    err_out("Can not set recv buf size:%s (%d)\n",
	    strerror(-rc),
	    -rc);
    return rc;
  }
  size=0;
  rc=internal_set_buffer(soc,SO_SNDBUF,_MSG_BUF_SIZE,_MSG_BUF_MIN_SIZE,&size);
  if (size < _MSG_BUF_MIN_SIZE) {
    err_out("Can not set send buf size:%s (%d)\n",
	    strerror(-rc),
	    -rc);
    return rc;
  }

  return 0;
}

int
setup_addr_info(struct addrinfo **infop, const char *ipaddr,int port,int stype,int family){
  int rc;
  int soc;
  struct addrinfo hints;
  struct addrinfo *res;
  struct addrinfo *node;
  char port_str[11];
  char host[NI_MAXHOST];
  char serv[NI_MAXSERV];

  if (!infop)
    return -EINVAL;

  memset(&hints,0,sizeof(struct addrinfo));
  snprintf(port_str,10,"%d",port);

  if (family)
    hints.ai_family=family;
  else
    hints.ai_family=PF_UNSPEC;
  
  hints.ai_socktype = stype;
  
  if (!ipaddr) {
    hints.ai_flags=AI_PASSIVE;
  }

  rc=getaddrinfo(ipaddr, port_str,&hints,&res);
  if (rc<0) {
    err_out("Can not get addr info:%s(%d)\n",gai_strerror(rc),rc);
    goto err_out;
  }
    
  for(node=res;node;node=node->ai_next) {
    rc=getnameinfo(res->ai_addr, res->ai_addrlen,
		host, sizeof(host),
		serv, sizeof(serv), 
		NI_NUMERICHOST|NI_NUMERICSERV);

    g_assert (!rc);

    dbg_out("host:%s service:%s (family,type,proto)=(%d,%d,%d)\n",
	    host,
	    serv,
	    res->ai_family,
	    res->ai_socktype,
	    res->ai_protocol);
  }
  *infop=res;

  return 0;
 err_out:
  return rc;
}

int
wait_socket(int soc,int wait_for,int sec) {
  fd_set rd_set;
  fd_set wr_set;
  int rc;
  struct timeval tmout_val;
  int i;

  for(i=0;i<NET_MAX_RETRY;++i) {
    FD_ZERO(&rd_set);
    FD_ZERO(&wr_set);

    if (wait_for & WAIT_FOR_READ)
      FD_SET(soc,&rd_set);
    if (wait_for & WAIT_FOR_WRITE)
      FD_SET(soc,&wr_set);

    dbg_out("select\n");
    tmout_val.tv_sec=sec;
    tmout_val.tv_usec=0;

    rc=select(soc+1,&rd_set,&wr_set,NULL,&tmout_val);
    if (rc>0)
      dbg_out("selectOK\n");
    if (rc<0) {
      if (errno == EINTR)
	continue;
      return -errno;
    }
    
    if ( (wait_for & WAIT_FOR_READ ) && (FD_ISSET(soc,&rd_set)) )
      return 0;
    if ( (wait_for & WAIT_FOR_WRITE ) && (FD_ISSET(soc,&wr_set)) )
      return 0;
  }
  return -EINVAL; 
}
int 
sock_set_recv_timeout(int soc,int sec){
  int rc;
  struct timeval tmout_val;  

  tmout_val.tv_sec=sec;
  tmout_val.tv_usec=0;
  dbg_out("Set timeout:%d\n",sec);

  rc=setsockopt(soc, SOL_SOCKET, SO_RCVTIMEO, (void*)&tmout_val, sizeof(struct timeval));
  if (rc<0) {
    err_out("Can not set timeout:%s(%d)\n",strerror(errno),errno);
    return -errno;
  }

  return 0;
}
