#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include "common.h"


#define MSG_STRING         "1:%ld:%s:%s:%lu:%s"
#define MSG_ABSENT_STRING  "1:%ld:%s:%s:%lu:%s[%s]"
#define MSG_STRING_RECVMSG "1:%ld:%s:%s:%lu:%d"
#define MSG_STRING_INFOMSG "1:%ld:%s:%s:%lu:%s"


static long pkt_no=0;

GStaticMutex pktno_mutex = G_STATIC_MUTEX_INIT;

long
ipmsg_get_pkt_no(void) {
  g_static_mutex_lock(&pktno_mutex);
  ++pkt_no;
  g_static_mutex_unlock(&pktno_mutex);
  return (time(NULL)%100000 + pkt_no);
}

static int
construct_message(const char *base,const char *ext,char **msg,size_t *len){
  char *sp;

  if ( (!msg) || (*msg) || (!len) )
    return -EINVAL;

  *len=strlen(base)+1;
  if (ext)
    *len += strlen(ext)+1;
  *msg=malloc(*len);
  if (!(*msg)) 
    return -ENOMEM;
  sp=*msg;

  memmove(sp,base,strlen(base)+1); /*ヌル終端をふくめてコピーする */
  sp += strlen(base)+1;
  if (ext)
    memmove(sp,ext,strlen(ext)+1); /*ヌル終端をふくめてコピーする */

  return 0;
}
static int
ipmsg_construct_file_attach_message(const char *msg_string,const char *ext_part,int flags,size_t *len,char **msg,long pktno){
  char base[IPMSG_BUFSIZ];
  char *message;
  size_t msg_len;
  int rc;


  if ( (!msg_string) || (!ext_part) || (!msg) || (!len) )
    return -EINVAL;
  if (!(flags & IPMSG_FILEATTACHOPT))
    return -EINVAL;

  dbg_out("ext:%s\n",ext_part);

  memset(base,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ,MSG_STRING,
	   pktno,
	   hostinfo_refer_user_name(),
	   hostinfo_refer_host_name(),
	   IPMSG_SENDMSG|flags,
	   msg_string);



  message=NULL;
  rc=construct_message(base,ext_part,&message,&msg_len);
  *len=msg_len;
  if (rc<0)
    return rc;

  *msg=message;


  return 0;
}
static int
ipmsg_construct_normal_message(const char *msg_string,int flags,size_t *len,char **msg,long pktno){
  char base[IPMSG_BUFSIZ];
  char *message;
  size_t msg_len;
  int rc;

  if ( (!msg_string) || (!msg) || (!len) )
    return -EINVAL;


  memset(base,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ,MSG_STRING,
	   pktno,
	   hostinfo_refer_user_name(),
	   hostinfo_refer_host_name(),
	   IPMSG_SENDMSG|flags,
	   msg_string);

  message=NULL;
  rc=construct_message(base,NULL,&message,&msg_len);

  if (rc<0)
    return rc;

  *msg=message;
  *len=msg_len;

  return 0;
}
int
ipmsg_construct_getfile_message(const char *msg_string,int flags,size_t *len,char **msg,long pktno){
  char base[IPMSG_BUFSIZ];
  char *message;
  size_t msg_len;
  int rc;

  if ( (!msg_string) || (!msg) || (!len) )
    return -EINVAL;


  memset(base,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ,MSG_STRING,
	   pktno,
	   hostinfo_refer_user_name(),
	   hostinfo_refer_host_name(),
	   IPMSG_GETFILEDATA|flags,
	   msg_string);

  message=NULL;
  rc=construct_message(base,NULL,&message,&msg_len);

  if (rc<0)
    return rc;

  *msg=message;
  *len=msg_len;

  return 0;
}

static int
ipmsg_send_absent_msg(const udp_con_t *con,const int flags){
  char *absent_message=NULL;
  gchar *absent_message_proto=NULL;
  char  *send_string;
  int index;
  size_t len;
  int local_flags;
  int rc;
  char *ipaddr;

  rc=-EINVAL;
  if (!con)
    goto no_need_free_out;

  local_flags=flags;

  g_assert(!hostinfo_get_absent_id(&index));
  rc=hostinfo_get_absent_message(index,(const char **)&absent_message);
  if (!absent_message)
    goto no_need_free_out;
  convert_string_ipmsg_proto(absent_message,(const gchar **)&absent_message_proto);
  if (!absent_message_proto) 
    goto free_absent_message;

  pkt_no=ipmsg_get_pkt_no();
  rc=ipmsg_construct_normal_message(absent_message_proto,flags,&len,&send_string,pkt_no);
  if (rc<0){
    if (send_string)
      goto free_send_string;
  }
  if (!send_string)
    goto free_absent_message;

  ipaddr=(char *)udp_get_peeraddr(con);

  dbg_out("send absent message to %s:%s\n",ipaddr,send_string);
  rc=register_sent_message(con,ipaddr,pkt_no,send_string,len);
  if (rc<0) {
    rc *= -1;
    err_out("Can not register message:%s (%d)\n",strerror(rc),rc);

  }
  rc=udp_send_peer(con,send_string,len);
  if (rc<0) {
    rc*=-1;
    err_out("Can not send: %s %d\n",strerror(rc),rc);
    goto free_send_string;
  }

  logfile_send_log(ipaddr,absent_message);
  rc=0;

free_send_string:
  free(send_string);
free_absent_message_proto:
  free (absent_message_proto);
free_absent_message:
  free(absent_message);
no_need_free_out:
  return rc;
}

int
ipmsg_send_br_entry(const udp_con_t *con,const int flags){
  char base[IPMSG_BUFSIZ];
  char ext[IPMSG_BUFSIZ];
  char *msg=NULL;
  size_t len;
  int rc;
  unsigned long local_flags;

  local_flags=IPMSG_BR_ENTRY|G2IPMSG_DEFAULT_ENTRY_FLAGS|flags;
  if (hostinfo_is_ipmsg_absent()) 
    local_flags |= IPMSG_ABSENCEOPT;

  memset(base,0,IPMSG_BUFSIZ);
  memset(ext,0,IPMSG_BUFSIZ);
  snprintf(base,IPMSG_BUFSIZ,MSG_STRING,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),local_flags,hostinfo_refer_nick_name_proto());
  snprintf(ext,IPMSG_BUFSIZ,"%s",hostinfo_refer_group_name_proto());
  if (hostinfo_is_ipmsg_absent()) {
    char *absent_title=NULL;
    gchar *absent_title_proto=NULL;
    int index;

    if (!hostinfo_get_absent_id(&index)) {
      hostinfo_get_absent_title(index,(const char **)&absent_title);
      g_assert (absent_title);
      convert_string_ipmsg_proto(absent_title,(const gchar **)&absent_title_proto);
      free(absent_title);
      g_assert (absent_title_proto);
      if (absent_title_proto) {
	snprintf(ext,IPMSG_BUFSIZ,"%s[%s]",hostinfo_refer_group_name_proto(),absent_title_proto);
	free (absent_title_proto);
      }
    }
  }

  rc=construct_message(base,ext,&msg,&len);
  if (rc<0)
    return rc;

  hostinfo_send_broad_cast(con,msg,len);
  free(msg);

  return 0;
}

int
ipmsg_send_br_exit(const udp_con_t *con,const int flags){
  char base[IPMSG_BUFSIZ];
  char ext[IPMSG_BUFSIZ];
  char *msg=NULL;
  size_t len;
  int rc;

  memset(base,0,IPMSG_BUFSIZ);
  memset(ext,0,IPMSG_BUFSIZ);
  snprintf(base,IPMSG_BUFSIZ,MSG_STRING,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),IPMSG_BR_EXIT|G2IPMSG_DEFAULT_ENTRY_FLAGS|flags,hostinfo_refer_nick_name_proto());
  snprintf(ext,IPMSG_BUFSIZ,"%s",hostinfo_refer_group_name_proto());

  rc=construct_message(base,ext,&msg,&len);
  if (rc<0)
    return rc;

  hostinfo_send_broad_cast(con,msg,len);
  free(msg);

  return 0;
}
int
ipmsg_send_br_absence(const udp_con_t *con,const int flags){
  char base[IPMSG_BUFSIZ];
  char ext[IPMSG_BUFSIZ];
  char *msg=NULL;
  size_t len;
  int rc;
  unsigned long local_flags;

  local_flags=IPMSG_BR_ABSENCE|G2IPMSG_DEFAULT_ENTRY_FLAGS|flags;
  if (hostinfo_is_ipmsg_absent()) 
    local_flags |= IPMSG_ABSENCEOPT;

  memset(base,0,IPMSG_BUFSIZ);
  memset(ext,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ-1,MSG_STRING,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),local_flags,hostinfo_refer_nick_name_proto());
  base[IPMSG_BUFSIZ-1]='\0';

  snprintf(ext,IPMSG_BUFSIZ,"%s",hostinfo_refer_group_name_proto());
  ext[IPMSG_BUFSIZ-1]='\0';

  if (hostinfo_is_ipmsg_absent()) {
    char *absent_title=NULL;
    gchar *absent_title_proto=NULL;
    int index;

    if (!hostinfo_get_absent_id(&index)) {
      hostinfo_get_absent_title(index,(const char **)&absent_title);
      g_assert (absent_title);
      convert_string_ipmsg_proto(absent_title,(const gchar **)&absent_title_proto);
      free(absent_title);
      g_assert (absent_title_proto);

      if (absent_title_proto) {
	memset(base,0,IPMSG_BUFSIZ);
	snprintf(base,IPMSG_BUFSIZ-1,MSG_ABSENT_STRING,
		 ipmsg_get_pkt_no(),
		 hostinfo_refer_user_name(),
		 hostinfo_refer_host_name(),
		 local_flags,
		 hostinfo_refer_nick_name_proto(),
		 absent_title_proto);
	base[IPMSG_BUFSIZ-1]='\0';
	free (absent_title_proto);
      }
    }
  }

  rc=construct_message(base,ext,&msg,&len);
  if (rc<0)
    return rc;

  hostinfo_send_broad_cast(con,msg,len);
  free(msg);

  return 0;
}

int
ipmsg_send_ans_entry(const udp_con_t *con,const int flags){
  char base[IPMSG_BUFSIZ];
  char ext[IPMSG_BUFSIZ];
  char *msg=NULL;
  size_t len;
  int rc;
  unsigned long local_flags;

  local_flags=IPMSG_ANSENTRY|G2IPMSG_DEFAULT_ENTRY_FLAGS |flags;
  if (hostinfo_is_ipmsg_absent()) 
    local_flags |= IPMSG_ABSENCEOPT;

  memset(base,0,IPMSG_BUFSIZ);
  memset(ext,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ,MSG_STRING,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),local_flags,hostinfo_refer_nick_name_proto());

  snprintf(ext,IPMSG_BUFSIZ,"%s",hostinfo_refer_group_name_proto());
  if (hostinfo_is_ipmsg_absent()) {
    char *absent_title=NULL;
    gchar *absent_title_proto=NULL;
    int index;

    if (!hostinfo_get_absent_id(&index)) {
      hostinfo_get_absent_title(index,(const char **)&absent_title);
      g_assert (absent_title);
      convert_string_ipmsg_proto(absent_title,(const gchar **)&absent_title_proto);
      free(absent_title);
      g_assert (absent_title_proto);
      if (absent_title_proto) {
	snprintf(ext,IPMSG_BUFSIZ,"%s[%s]",hostinfo_refer_group_name_proto(),absent_title_proto);
	free (absent_title_proto);
      }
    }
  }

  rc=construct_message(base,ext,&msg,&len);
  if (rc<0)
    return rc;
  dbg_out("AnserEntry:%s\n",msg);
  udp_send_peer(con,msg,len);
  free(msg);

  return 0;
}
int
ipmsg_send_recv_msg(const udp_con_t *con,const msg_data_t *orig_msg){
  char base[IPMSG_BUFSIZ];
  char *msg=NULL;
  size_t len;
  int rc;

  memset(base,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ,MSG_STRING_RECVMSG,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),IPMSG_RECVMSG,refer_pkt_no_name_from_msg(orig_msg));

  rc=construct_message(base,NULL,&msg,&len);
  if (rc<0)
    return rc;
  udp_send_peer(con,msg,len);
  free(msg);

  return 0;
}
int
ipmsg_send_release_files(const udp_con_t *con,const char *ipaddr,int pktno){
  char base[IPMSG_BUFSIZ];
  char *msg=NULL;
  size_t len;
  int rc;

  memset(base,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ,MSG_STRING_RECVMSG,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),IPMSG_RELEASEFILES,pktno);

  rc=construct_message(base,NULL,&msg,&len);
  if (rc<0)
    return rc;
  rc=udp_send_message(con,ipaddr,hostinfo_refer_ipmsg_port(),msg,len);
  free(msg);

  if (rc<0)
    return rc;

  return 0;
}

int
ipmsg_send_read_msg(const udp_con_t *con,const char *ipaddr,int pktno){
  char base[IPMSG_BUFSIZ];
  char *msg=NULL;
  size_t len;
  int rc;

  memset(base,0,IPMSG_BUFSIZ);

  snprintf(base,IPMSG_BUFSIZ,MSG_STRING_RECVMSG,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),IPMSG_READMSG,pktno);

  rc=construct_message(base,NULL,&msg,&len);
  if (rc<0)
    return rc;
  rc=udp_send_message(con,ipaddr,hostinfo_refer_ipmsg_port(),msg,len);
  free(msg);

  if (rc<0)
    return rc;

  return 0;
}

int
ipmsg_send_send_msg(const udp_con_t *con,const char *ipaddr,int flags,int pkt_no,const char *message,const char *ext_part){
  int rc;
  size_t len;
  char *send_string=NULL;
  gchar *internal_message=NULL;

  if ( (!con) || (!ipaddr) || (!message) )
    return -EINVAL;
  if ( (flags & IPMSG_FILEATTACHOPT) && (!ext_part) )
    return -EINVAL;
    
  if (flags & IPMSG_FILEATTACHOPT)
    rc=ipmsg_construct_file_attach_message(message,ext_part,flags,&len,&send_string,pkt_no);
  else
    rc=ipmsg_construct_normal_message(message,flags,&len,&send_string,pkt_no);
  if (rc<0){
    if (send_string)
      free(send_string);
    return rc;
  }
  dbg_out("send message to %s:%s\n",ipaddr,send_string);
  rc=register_sent_message(con,ipaddr,pkt_no,send_string,len);
  if (rc<0) {
    rc *= -1;
    err_out("Can not register message:%s (%d)\n",strerror(rc),rc);
    return -rc;
  }

  rc=udp_send_message(con,ipaddr,hostinfo_refer_ipmsg_port(), send_string,len);
  if (rc<0) {
    rc*=-1;
    err_out("Can not send: %s %d\n",strerror(rc),rc);
    goto err_out;
  }
  convert_string_internal(message,(const gchar **)&internal_message);
  if (internal_message) {
    logfile_send_log(ipaddr,internal_message);
    g_free(internal_message);
  }

  rc=0;
 err_out:
  free(send_string);
  return rc;
}
int
ipmsg_proc_br_absence(const udp_con_t *con,const msg_data_t *msg){
  dbg_out("\nNew user: \n\tUser:%s Host: %s\n\tNickName: %s GroupName:%s\n",
	  refer_user_name_from_msg(msg),
	  refer_host_name_from_msg(msg),
	  refer_nick_name_from_msg(msg),
	  refer_group_name_from_msg(msg));

  userdb_update_user(con,msg);
  userdb_print_user_list();

  return 0;
}
int
ipmsg_proc_br_entry(const udp_con_t *con,const msg_data_t *msg) {
  int rc=0;

  ipmsg_send_ans_entry(con,0);

  rc=userdb_add_user(con,msg);
  if (rc == -EEXIST)
    rc=userdb_update_user(con,msg);
  userdb_print_user_list();

  return rc;
}
int
ipmsg_proc_br_exit(const udp_con_t *con,const msg_data_t *msg) {
  dbg_out("\nLeave user: \n\tUser:%s Host: %s\n\tNickName: %s GroupName:%s\n",
	  refer_user_name_from_msg(msg),
	  refer_host_name_from_msg(msg),
	  refer_nick_name_from_msg(msg),
	  refer_group_name_from_msg(msg));
  userdb_del_user(con,msg);
  userdb_print_user_list();

  return 0;
}
int
ipmsg_proc_ans_entry(const udp_con_t *con,const msg_data_t *msg) {
  dbg_out("\nNew user: \n\tUser:%s Host: %s\n\tNickName: %s GroupName:%s\n",
	  refer_user_name_from_msg(msg),
	  refer_host_name_from_msg(msg),
	  refer_nick_name_from_msg(msg),
	  refer_group_name_from_msg(msg));
  userdb_add_user(con,msg);
  userdb_print_user_list();

  return 0;
}

int
ipmsg_proc_send_msg(const udp_con_t *con,const msg_data_t *msg) {
  gchar *internal_message=NULL;
  int flags;

  if (msg->command_opts & IPMSG_SENDCHECKOPT) {
    dbg_out("Send check opt is enabled\n");
    ipmsg_send_recv_msg(con,msg);
  }
  dbg_out("\nmessage:\n%s\n",msg->message);

  convert_string_internal(msg->message,(const gchar **)&internal_message);
  if (internal_message) {
    logfile_recv_log(udp_get_peeraddr(con),internal_message);
    g_free(internal_message);
  }

  if (hostinfo_is_ipmsg_absent()){
    flags=hostinfo_get_normal_send_flags();
    if (hostinfo_refer_ipmsg_default_secret())
      flags |= IPMSG_SECRETOPT;
    ipmsg_send_absent_msg(con,flags);
  }

  if (hostinfo_refer_ipmsg_default_popup()) /* if it is set, should not pop-up the win */
    store_message_window(msg,udp_get_peeraddr(con));
  else
    recv_message_window(msg,udp_get_peeraddr(con));

  return 0;
}
int
ipmsg_proc_read_msg(const udp_con_t *con,const msg_data_t *msg) {
  long pktno;
  gchar *user;
  struct timeval tv;

  gettimeofday(&tv, NULL);
  pktno=strtol(msg->message, (char **)NULL, 10);
  dbg_out("read mssage:seq %ld\n",pkt_no);
  /* TODO:開封確認 */
  dbg_out("read mssage:seq %s\n",msg->message);
  user=strdup(refer_user_name_from_msg(msg));
  /*  メモリ不足の場合はNULLを送信。処理は受け手で考える  */
  read_message_dialog(user,udp_get_peeraddr(con),tv.tv_sec);

  return 0;
}
int
ipmsg_proc_release_files_msg(const udp_con_t *con,const msg_data_t *msg) {
  int pktno;
  int rc;


  pktno=(int)strtol(msg->message, (char **)NULL, 10);
  dbg_out("release files mssage:seq %d(%s)\n",pktno,msg->message);
  rc=release_attach_file_block(pktno,FALSE);
  if (rc<0)
    err_out("Can not release attach file information:pktno=%d %s (%d)\n",pktno,strerror(-rc),-rc);
  return rc;
}
int
ipmsg_proc_recv_msg(const udp_con_t *con,const msg_data_t *msg) {
  long pktno;

  if ( (!msg) ||  (!(msg->message)))
    return -EINVAL;

  pktno=(long)strtol(msg->message, (char **)NULL, 10);
  dbg_out("recv mssage:seq %ld\n",pkt_no);
  unregister_sent_message(pktno);
  return 0;
}
static int
ipmsg_proc_send_absence_info(const udp_con_t *con,const msg_data_t *orig_msg){
  char base[IPMSG_BUFSIZ];
  char *absent_message=NULL;
  gchar *absent_message_proto=NULL;
  int index;
  char *msg=NULL;
  size_t len;
  int rc;

  dbg_out("here\n");

  memset(base,0,IPMSG_BUFSIZ);

  if (hostinfo_is_ipmsg_absent()) {
    if (!hostinfo_get_absent_id(&index)) {
      hostinfo_get_absent_message(index,(const char **)&absent_message);
      g_assert (absent_message);
    }
    convert_string_ipmsg_proto(absent_message,(const gchar **)&absent_message_proto);
    free(absent_message);
  }else{
    absent_message=_("I am here.");
    convert_string_ipmsg_proto(absent_message,(const gchar **)&absent_message_proto);
  }
  g_assert (absent_message_proto);
  
  if (absent_message_proto) {
    snprintf(base,IPMSG_BUFSIZ,MSG_STRING_INFOMSG,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),(IPMSG_AUTORETOPT|IPMSG_SENDABSENCEINFO),absent_message_proto);
    free (absent_message_proto);
  }

  rc=construct_message(base,NULL,&msg,&len);
  if (rc<0)
    return rc;
  udp_send_peer(con,msg,len);
  free(msg);

  return 0;
}
static int
ipmsg_proc_send_version_info(const udp_con_t *con,const msg_data_t *orig_msg){
  char base[IPMSG_BUFSIZ];
  char ver[IPMSG_BUFSIZ];
  int index;
  char *msg=NULL;
  size_t len;
  int rc;

  dbg_out("here\n");

  memset(base,0,IPMSG_BUFSIZ);
  memset(ver,0,IPMSG_BUFSIZ);

  snprintf(ver,IPMSG_BUFSIZ,"GNOME2 %s (%s) %s:%s",_("Edition"),PACKAGE,_("Version"),VERSION);
  snprintf(base,IPMSG_BUFSIZ,MSG_STRING_INFOMSG,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),(IPMSG_AUTORETOPT|IPMSG_SENDINFO),ver);

  rc=construct_message(base,NULL,&msg,&len);
  if (rc<0)
    return rc;
  udp_send_peer(con,msg,len);
  free(msg);

  return 0;
}
static int
ipmsg_proc_br_isgetlist2(const udp_con_t *con,const msg_data_t *msg){
  char base[IPMSG_BUFSIZ];
  char ver[IPMSG_BUFSIZ];
  size_t len;
  char *message=NULL;
  int rc;

  dbg_out("here\n");

  if ( (!con) || (!msg) )
    return -EINVAL;


  memset(base,0,IPMSG_BUFSIZ);
  snprintf(base,IPMSG_BUFSIZ,MSG_STRING_INFOMSG,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),(IPMSG_AUTORETOPT|IPMSG_OKGETLIST),"");

  rc=construct_message(base,NULL,&message,&len);

  if (rc<0)
    return rc;

  udp_send_peer(con,message,len);
  free(message);

  return 0;
}
static int
ipmsg_proc_getlist(const udp_con_t *con,const msg_data_t *msg){
  char base[IPMSG_BUFSIZ];
  char ver[IPMSG_BUFSIZ];
  int length;
  char *string=NULL;
  int flags=0;
  int index;
  int start;
  char *message=NULL;
  size_t len;
  int rc;

  dbg_out("here\n");

  if ( (!con) || (!msg) )
    return -EINVAL;

  length=G2IPMSG_DEFAULT_HOST_LIST_LEN;

  start=0;
  start=strtol(msg->message, (char **)NULL, 10);
  userdb_get_hostlist_string(start,&length, &string,&flags);
  dbg_out("return:%s\n",string);
  memset(base,0,IPMSG_BUFSIZ);
  snprintf(base,IPMSG_BUFSIZ,MSG_STRING_INFOMSG,ipmsg_get_pkt_no(),hostinfo_refer_user_name(),hostinfo_refer_host_name(),(IPMSG_AUTORETOPT|IPMSG_ANSLIST),string);
  if (string)
    free(string);

  rc=construct_message(base,NULL,&message,&len);
  dbg_out("send:%s\n",message);
  if (rc<0)
    return rc;

  udp_send_peer(con,message,len);
  dbg_out("return:%s\n",message);
  free(message);

  return 0;
}

int
ipmsg_dispatch_message(const udp_con_t *con,const msg_data_t *msg){
  
  if ( (!con) || (!msg) )
    return -EINVAL;

  switch(msg->command) {
  case IPMSG_NOOPERATION:
    break;
  case IPMSG_BR_ENTRY:
    dbg_out("Dispatch br_entry\n");
    ipmsg_proc_br_entry(con,msg);
    break;
  case IPMSG_BR_EXIT:
    dbg_out("Dispatch br_exit\n");
    ipmsg_proc_br_exit(con,msg);
    break;
  case IPMSG_ANSENTRY:
    dbg_out("Dispatch ans_entry\n");
    ipmsg_proc_ans_entry(con,msg);
    break;
  case IPMSG_BR_ABSENCE:
    dbg_out("Dispatch br_absense\n");
    ipmsg_proc_br_absence(con,msg);
    break;
  case IPMSG_BR_ISGETLIST:
    dbg_out("Dispatch ok_isget_list\n");
    break;
  case IPMSG_OKGETLIST:
    dbg_out("Dispatch ok_get_list\n");
    break;
  case IPMSG_GETLIST:
    dbg_out("Dispatch get_list\n");
#if 0
    ipmsg_proc_getlist(con,msg);
#endif
    break;
  case IPMSG_ANSLIST:
    dbg_out("Dispatch br_ans_list\n");
    break;
  case IPMSG_BR_ISGETLIST2:
    dbg_out("Dispatch br_isget_list2\n");
#if 0
    ipmsg_proc_br_isgetlist2(con,msg);
#endif
    break;
  case IPMSG_SENDMSG:
    dbg_out("Dispatch send_message\n");
    ipmsg_proc_send_msg(con,msg);
    break;
  case IPMSG_RECVMSG:
    dbg_out("Dispatch recv_message\n");
    ipmsg_proc_recv_msg(con,msg);
    break;
  case IPMSG_READMSG:
    dbg_out("Dispatch read_message\n");
    ipmsg_proc_read_msg(con,msg);
    break;
  case IPMSG_DELMSG:
    dbg_out("Dispatch delete_message\n");
    break;
  case IPMSG_ANSREADMSG:
    dbg_out("Dispatch ans_read_message\n");
    break;
  case IPMSG_GETINFO:
    dbg_out("Dispatch get_info\n");
    ipmsg_proc_send_version_info(con,msg);
    break;
  case IPMSG_SENDINFO:
    dbg_out("Dispatch send_info\n");
    break;
  case IPMSG_GETABSENCEINFO:
    dbg_out("Dispatch get_absence_info\n");
    ipmsg_proc_send_absence_info(con,msg);
    break;
  case IPMSG_SENDABSENCEINFO:
    dbg_out("Dispatch send_absence_info\n");
    break;
  case IPMSG_RELEASEFILES:
    dbg_out("Dispatch release files\n");
    ipmsg_proc_release_files_msg(con,msg);
    break;
  default:
    dbg_out("Can not dispatch unknown:%x\n",msg->command);
    /* Unknown command */
    break;
  }
  return 0;
}
